import logging
from abc import ABCMeta
from typing import Any, List, Type, Tuple, Union, TypeVar, Optional, cast

from dependency_injector.wiring import Provide, inject

from qm.type_hinting import Value
from qm.exceptions import QmValueError
from qm.utils.async_utils import run_async
from qm.grpc.general_messages import Matrix
from qm.api.models.jobs import PendingJobData
from qm.grpc.qm_manager import GetRunningJobRequest
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.utils.general_utils import create_input_stream_name
from qm.api.base_api import BaseApi, connection_error_handle
from qm.containers.capabilities_container import CapabilitiesContainer
from qm.api.stubs.deprecated_job_manager_stub import DeprecatedJobManagerServiceStub
from qm.grpc.job_manager import (
    IntStreamData,
    BoolStreamData,
    FixedStreamData,
    JobManagerServiceStub,
    InsertInputStreamRequest,
    GetElementCorrectionRequest,
    SetElementCorrectionRequest,
)
from qm._QmJobErrors import (
    MissingJobError,
    MissingElementError,
    UnknownInputStreamError,
    ElementWithSingleInputError,
    InvalidElementCorrectionError,
    InvalidJobExecutionStatusError,
    ElementWithoutIntermediateFrequencyError,
    _handle_job_manager_error,
)
from qm.grpc.frontend import (
    HaltRequest,
    FrontendStub,
    ResumeRequest,
    JobQueryParams,
    QueryValueMatcher,
    JobExecutionStatus,
    IsJobRunningRequest,
    PausedStatusRequest,
    IsJobAcquiringDataRequest,
    GetJobExecutionStatusRequest,
    IsJobAcquiringDataResponseAcquiringStatus,
)

logger = logging.getLogger(__name__)


JobStubType = TypeVar("JobStubType", JobManagerServiceStub, DeprecatedJobManagerServiceStub)


@connection_error_handle()
class JobManagerBaseApi(BaseApi[JobStubType], metaclass=ABCMeta):
    def __init__(self, connection_details: ConnectionDetails):
        super().__init__(connection_details)
        self._frontend_stub = FrontendStub(self._channel)

    def set_element_correction(
        self, job_id: str, element_name: str, correction: Matrix
    ) -> Tuple[float, float, float, float]:
        request = SetElementCorrectionRequest(job_id=job_id, qe_name=element_name, correction=correction)

        response = run_async(self._stub.set_element_correction(request, timeout=self._timeout))

        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            InvalidElementCorrectionError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def get_element_correction(self, job_id: str, element_name: str) -> Tuple[float, float, float, float]:
        request = GetElementCorrectionRequest(job_id=job_id, qe_name=element_name)

        response = run_async(self._stub.get_element_correction(request, timeout=self._timeout))
        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def insert_input_stream(self, job_id: str, stream_name: str, data: List[Value]) -> None:
        stream_name = create_input_stream_name(stream_name)
        request = InsertInputStreamRequest(job_id=job_id, stream_name=stream_name)

        if all(type(element) == bool for element in data):
            request.bool_stream_data = BoolStreamData(data=cast(List[bool], data))
        elif all(type(element) == int for element in data):
            request.int_stream_data = IntStreamData(data=cast(List[int], data))
        elif all(type(element) == float for element in data):
            request.fixed_stream_data = FixedStreamData(data=cast(List[float], data))
        else:
            raise QmValueError(
                f"Invalid type in data, type is '{set(type(el) for el in data)}', "
                f"excepted types are bool | int | float"
            )

        response = run_async(self._stub.insert_input_stream(request, timeout=self._timeout))

        valid_errors = (
            MissingJobError,
            InvalidJobExecutionStatusError,
            UnknownInputStreamError,
        )
        _handle_job_manager_error(request, response, valid_errors)

    def halt(self, job_id: str) -> bool:
        request = HaltRequest(job_id=job_id)
        response = run_async(self._frontend_stub.halt(request, timeout=self._timeout))
        return response.ok

    def resume(self, job_id: str) -> bool:
        request = ResumeRequest(job_id=job_id)
        run_async(self._frontend_stub.resume(request, timeout=self._timeout))
        return True

    def is_paused(self, job_id: str) -> bool:
        request = PausedStatusRequest(job_id=job_id)
        response = run_async(self._frontend_stub.paused_status(request, timeout=self._timeout))
        return response.is_paused

    def is_job_running(self, job_id: str) -> bool:
        request = IsJobRunningRequest(job_id=job_id)

        response = run_async(self._frontend_stub.is_job_running(request, timeout=self._timeout))
        return response.is_running

    def is_data_acquiring(self, job_id: str) -> IsJobAcquiringDataResponseAcquiringStatus:
        request = IsJobAcquiringDataRequest(job_id=job_id)
        response = run_async(self._frontend_stub.is_job_acquiring_data(request, timeout=self._timeout))
        return response.acquiring_status

    def get_job_execution_status(self, job_id: str, quantum_machine_id: str) -> JobExecutionStatus:
        request = GetJobExecutionStatusRequest(job_id=job_id, quantum_machine_id=quantum_machine_id)
        response = run_async(self._frontend_stub.get_job_execution_status(request, timeout=self._timeout))
        return response.status

    @staticmethod
    def _create_job_query_params(
        quantum_machine_id: str,
        job_id: Optional[str],
        position: Optional[int],
        user_id: Optional[str],
    ) -> JobQueryParams:
        request = JobQueryParams(quantum_machine_id=quantum_machine_id)

        if position is not None:
            request.position = position

        if job_id is not None:
            request.job_id = QueryValueMatcher(value=job_id)

        if user_id is not None:
            request.user_id = QueryValueMatcher(value=user_id)

        return request

    def remove_job(
        self,
        quantum_machine_id: str,
        job_id: Optional[str] = None,
        position: Optional[int] = None,
        user_id: Optional[str] = None,
    ) -> int:
        request = JobManagerBaseApi._create_job_query_params(quantum_machine_id, job_id, position, user_id)

        response = run_async(self._frontend_stub.remove_pending_jobs(request, timeout=self._timeout))
        return response.numbers_of_jobs_removed

    def get_pending_jobs(
        self,
        quantum_machine_id: str,
        job_id: Optional[str],
        position: Optional[int],
        user_id: Optional[str],
    ) -> List[PendingJobData]:
        request = JobManagerBaseApi._create_job_query_params(quantum_machine_id, job_id, position, user_id)
        response = run_async(self._frontend_stub.get_pending_jobs(request, timeout=self._timeout))
        return [
            PendingJobData(
                job_id=job_id,
                position_in_queue=status.position_in_queue,
                time_added=status.time_added,
                added_by=status.added_by,
            )
            for job_id, status in response.pending_jobs.items()
        ]

    def get_running_job(self, machine_id: str) -> Optional[str]:
        request = GetRunningJobRequest(machine_id=machine_id)
        response = run_async(self._frontend_stub.get_running_job(request, timeout=self._timeout))

        if response.job_id:
            return response.job_id
        return None


class JobManagerApi(JobManagerBaseApi[JobManagerServiceStub]):
    @property
    def _stub_class(self) -> Type[JobManagerServiceStub]:
        return JobManagerServiceStub


class DeprecatedJobManagerApi(JobManagerBaseApi[DeprecatedJobManagerServiceStub]):
    @property
    def _stub_class(self) -> Type[DeprecatedJobManagerServiceStub]:
        return DeprecatedJobManagerServiceStub


@inject
def create_job_manager_from_api(
    api: BaseApi[Any], capabilities: ServerCapabilities = Provide[CapabilitiesContainer.capabilities]
) -> Union[JobManagerApi, DeprecatedJobManagerApi]:
    if capabilities.supports_new_grpc_structure:
        return JobManagerApi(api.connection_details)
    return DeprecatedJobManagerApi(api.connection_details)



from typing import List, Generic, TypeVar

T = TypeVar("T")


class Chunk(Generic[T]):
    def __init__(self) -> None:
        self._data: List[T] = []
        self._accepts_different = True

    def add(self, element: T) -> None:
        if self._data and element == self._data[-1]:
            self._data.append(element)
            self._accepts_different = False
            return

        if self.accepts_different:
            self._data.append(element)
            return

        raise ValueError("Tried to add number to a chunk that is already uniform")

    def __len__(self) -> int:
        return len(self._data)

    @property
    def data(self) -> List[T]:
        return self._data

    @property
    def first(self) -> T:
        return self._data[0]

    @property
    def accepts_different(self) -> bool:
        return self._accepts_different

    def __str__(self) -> str:
        if self.accepts_different:
            return str(self._data)
        return f"[{self.first}] * {len(self)}"


def split_list_to_chunks(list_data: List[T]) -> List[Chunk[T]]:
    curr_chunk: Chunk[T] = Chunk()
    chunks = [curr_chunk]
    for idx, curr_item in enumerate(list_data):
        if idx >= 1 and curr_item != list_data[idx - 1]:
            item_equals_next = idx < len(list_data) - 1 and curr_item == list_data[idx + 1]
            if item_equals_next or not curr_chunk.accepts_different:
                curr_chunk = Chunk()
                chunks.append(curr_chunk)

        curr_chunk.add(curr_item)
    return chunks



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: qm/pb/qm_manager.proto
# plugin: python-betterproto
# This file has been @generated

from dataclasses import dataclass
from typing import (
    List,
    Optional,
)

import betterproto

from .. import (
    general_messages as _general_messages__,
    qua_config as _qua_config__,
)


@dataclass(eq=False, repr=False)
class OpenQuantumMachineRequest(betterproto.Message):
    config: "_qua_config__.QuaConfig" = betterproto.message_field(1)
    never: bool = betterproto.bool_field(2, group="oneOfCloseOtherMachines")
    always: bool = betterproto.bool_field(3, group="oneOfCloseOtherMachines")
    if_needed: bool = betterproto.bool_field(4, group="oneOfCloseOtherMachines")
    keep_dc_offsets_when_closing: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class OpenQuantumMachineResponse(betterproto.Message):
    machine_id: str = betterproto.string_field(1)
    success: bool = betterproto.bool_field(2)
    config_validation_errors: List[
        "ConfigValidationMessage"
    ] = betterproto.message_field(3)
    physical_validation_errors: List[
        "PhysicalValidationMessage"
    ] = betterproto.message_field(4)
    open_qm_warnings: List["OpenQmWarning"] = betterproto.message_field(5)


@dataclass(eq=False, repr=False)
class CloseQuantumMachineRequest(betterproto.Message):
    machine_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CloseQuantumMachineResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GetQuantumMachineRequest(betterproto.Message):
    machine_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetQuantumMachineResponse(betterproto.Message):
    machine_id: str = betterproto.string_field(1)
    config: "_qua_config__.QuaConfig" = betterproto.message_field(2)
    success: bool = betterproto.bool_field(3)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class GetRunningJobRequest(betterproto.Message):
    machine_id: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class GetRunningJobResponse(betterproto.Message):
    machine_id: str = betterproto.string_field(1)
    job_id: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ListOpenQuantumMachinesResponse(betterproto.Message):
    machine_i_ds: List[str] = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CloseAllQuantumMachinesResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)
    errors: List["_general_messages__.ErrorMessage"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class GetControllersResponse(betterproto.Message):
    controllers: List["Controller"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Controller(betterproto.Message):
    name: str = betterproto.string_field(1)
    temperature: Optional[float] = betterproto.double_field(2, optional=True)


@dataclass(eq=False, repr=False)
class ConfigValidationMessage(betterproto.Message):
    message: str = betterproto.string_field(1)
    group: str = betterproto.string_field(2)
    path: str = betterproto.string_field(3)
    level: "_general_messages__.MessageLevel" = betterproto.enum_field(4)


@dataclass(eq=False, repr=False)
class PhysicalValidationMessage(betterproto.Message):
    message: str = betterproto.string_field(1)
    group: str = betterproto.string_field(2)
    path: str = betterproto.string_field(3)
    level: "_general_messages__.MessageLevel" = betterproto.enum_field(4)


@dataclass(eq=False, repr=False)
class OpenQmWarning(betterproto.Message):
    code: int = betterproto.int32_field(1)
    message: str = betterproto.string_field(2)



from abc import ABCMeta
from typing import Any, TypeVar, Protocol, Coroutine

import betterproto

from qm.utils.async_utils import run_async
from qm.exceptions import ErrorType, QopResponseError
from qm.api.base_api import BaseApi, StubType, connection_error_handle_decorator

SuccessType = TypeVar("SuccessType", bound=betterproto.Message)


class ResponseProtocol(Protocol[SuccessType, ErrorType]):
    success: SuccessType
    error: ErrorType


class BaseApiV2(BaseApi[StubType], metaclass=ABCMeta):
    @staticmethod
    @connection_error_handle_decorator
    def _run(coroutine: Coroutine[Any, Any, ResponseProtocol[SuccessType, ErrorType]]) -> SuccessType:
        response = run_async(coroutine)
        try:  # This one replaces the serialized-on-wire method, while we don't know the group name of the object
            return response.success
        except AttributeError:
            raise QopResponseError(error=response.error)



from pprint import pformat
from collections import defaultdict
from collections.abc import Collection
from typing import Any, List, Tuple, Generic, TypeVar

import betterproto
from marshmallow import ValidationError

from qm.StreamMetadata import StreamMetadataError


class QmQuaException(Exception):
    def __init__(self, message: str, *args: Any):
        self.message = message
        super().__init__(message, *args)


class QmmException(QmQuaException):
    pass


class OctaveConfigDeprecationException(QmmException):
    def __init__(self) -> None:
        super().__init__(
            "Received a response from the QM-app that includes octaves, please remove the OctaveConfig "
            "and move the configuration to the QUA-config."
        )


class OpenQmException(QmQuaException):
    def __init__(self, message: str, *args: Any, errors: List[Tuple[str, str, str]]):
        super().__init__(message, *args)
        self.errors = errors


class FailedToExecuteJobException(QmQuaException):
    pass


class FailedToAddJobToQueueException(QmQuaException):
    pass


class CompilationException(QmQuaException):
    pass


class JobCancelledError(QmQuaException):
    pass


class ErrorJobStateError(QmQuaException):
    def __init__(self, *args: Any, error_list: List[str]):
        super().__init__(*args)
        self._error_list = error_list if error_list else []

    def __str__(self) -> str:
        errors_string = "\n".join(error for error in self._error_list)
        return f"{super().__str__()}\n{errors_string}"


class UnknownJobStateError(QmQuaException):
    pass


class InvalidStreamMetadataError(QmQuaException):
    def __init__(self, stream_metadata_errors: List[StreamMetadataError], *args: Any):
        stream_errors_message = "\n".join(f"{e.error} at: {e.location}" for e in stream_metadata_errors)
        message = f"Error creating stream metadata:\n{stream_errors_message}"
        super().__init__(message, *args)


class ConfigValidationException(QmQuaException):
    pass


def _format_validation_error(curr_error: object) -> object:
    if isinstance(curr_error, defaultdict):
        to_return = {}
        for k, v in curr_error.items():
            for error in v.values():
                to_return[k] = _format_validation_error(error)
                # There is a hidden assumption here of a single error per entry
                break
        return to_return
    elif isinstance(curr_error, dict):
        to_return = {}
        for k, v in curr_error.items():
            to_return[k] = _format_validation_error(v)
        return to_return
    elif isinstance(curr_error, Collection):
        if all(isinstance(i, str) for i in curr_error):
            return curr_error
        else:
            return [_format_validation_error(i) for i in curr_error]
    raise ValueError(f"Unexpected type {type(curr_error)} in validation error")


class ConfigSchemaError(ConfigValidationException):
    def __init__(self, data: ValidationError) -> None:
        self.data = data
        self.formatted_dict = _format_validation_error(data.messages)
        super().__init__(pformat(self.formatted_dict, width=120))


class NoInputsOrOutputsError(ConfigValidationException):
    def __init__(self) -> None:
        super().__init__("An element must have either outputs or inputs. Please specify at least one.")


class ConfigSerializationException(QmQuaException):
    pass


class UnsupportedCapabilityError(QmQuaException):
    pass


class InvalidConfigError(QmQuaException):
    pass


class QMHealthCheckError(QmQuaException):
    pass


class QMFailedToGetQuantumMachineError(QmQuaException):
    pass


class QMSimulationError(QmQuaException):
    pass


class QmFailedToCloseQuantumMachineError(QmQuaException):
    pass


class QMFailedToCloseAllQuantumMachinesError(QmFailedToCloseQuantumMachineError):
    pass


class QMRequestError(QmQuaException):
    pass


class QMConnectionError(QmQuaException):
    pass


class QMTimeoutError(QmQuaException):
    pass


class QMRequestDataError(QmQuaException):
    pass


class QmServerDetectionError(QmQuaException):
    pass


class QmValueError(QmQuaException, ValueError):
    pass


class QmInvalidSchemaError(QmQuaException):
    pass


class QmInvalidResult(QmQuaException):
    pass


class QmNoResultsError(QmQuaException):
    pass


class FunctionInputError(QmQuaException):
    pass


class AnotherJobIsRunning(QmQuaException):
    def __init__(self) -> None:
        super().__init__("Another job is running on the QM. Halt it first")


class CalibrationException(QmQuaException):
    pass


class CantCalibrateElementError(CalibrationException):
    pass


class OctaveConnectionError(QmQuaException):
    pass


class OctaveLoopbackError(OctaveConnectionError):
    def __init__(self) -> None:
        super().__init__("lo loopback between different octave devices are not supported.")


class NoOutputPortDeclared(OctaveConnectionError):
    pass


class OctaveCableSwapError(OctaveConnectionError):
    def __init__(self) -> None:
        super().__init__("Cable swap detected. Please check your connections.")


class ElementUpconverterDeclarationError(OctaveConnectionError):
    def __init__(self) -> None:
        super().__init__(
            "Element declaration error, the I and Q connections are not connected to the same upconverter."
        )


class LOFrequencyMismatch(ConfigValidationException):
    def __init__(self) -> None:
        super().__init__(
            "LO frequency mismatch. The frequency stated in the element is different from "
            "the one stated in the Octave, remove the one in the element."
        )


class OctaveConnectionAmbiguity(ConfigValidationException):
    def __init__(self) -> None:
        super().__init__(
            "It is not allowed to override the default connection of the Octave. You should either state the "
            "default connection to Octave in the controller level, or set each port separately in the port level."
        )


class InvalidOctaveParameter(ConfigValidationException):
    pass


class ElementOutputConnectionAmbiguity(ConfigValidationException):
    pass


class QmRedirectionError(QmQuaException):
    pass


class QmLocationParsingError(QmQuaException):
    pass


class ElementInputConnectionAmbiguity(ConfigValidationException):
    pass


class StreamProcessingDataLossError(QmQuaException):
    pass


ErrorType = TypeVar("ErrorType", bound=betterproto.Message)


class QopResponseError(Exception, Generic[ErrorType]):
    def __init__(self, error: ErrorType):
        self._error = error

    @property
    def error(self) -> ErrorType:
        return self._error

    def __str__(self) -> str:
        details = getattr(self._error, "details", "")
        to_return = f"Error from QOP, details:\n{details}." if details else f"Error from QOP: {self._error}"
        return to_return



from collections.abc import Iterable as IterableClass
from typing import (
    Any,
    Set,
    Type,
    Union,
    TypeVar,
    Iterable,
    Sequence,
    Collection,
    SupportsInt,
    SupportsFloat,
    SupportsIndex,
    cast,
)

import numpy as np
import numpy.typing

from qm.exceptions import QmValueError
from qm.type_hinting import Value, NumberT

GeneralConversionType = Union[str, bytes, bytearray, memoryview]
FloatConversionType = Union[SupportsFloat, SupportsIndex, GeneralConversionType]
IntConversionType = Union[SupportsInt, SupportsIndex, GeneralConversionType]
Bool = Union[bool, np.bool_]

ConversionType = TypeVar("ConversionType", IntConversionType, FloatConversionType, Bool)


def convert_object_type(obj: ConversionType) -> Value:
    if isinstance(obj, (np.bool_, bool)):
        return bool(obj)
    if isinstance(obj, (np.integer, int)):
        return int(obj)
    if isinstance(obj, (np.floating, float)):
        return float(obj)
    raise QmValueError(f"cannot convert {type(obj)} to int | float | bool")


def get_all_iterable_data_types(it: Iterable[Any]) -> Set[Type[Any]]:
    return {type(e) for e in it}


NumberType = TypeVar("NumberType")


def collection_has_type(
    collection: Collection[NumberType], type_to_check: Type[NumberType], include_subclasses: bool
) -> bool:
    if include_subclasses:
        return any(isinstance(i, type_to_check) for i in collection)
    else:
        return any(type(i) is type_to_check for i in collection)


def collection_has_type_bool(collection: Collection[NumberType]) -> bool:
    return collection_has_type(collection, bool, False) or collection_has_type(collection, np.bool_, True)


def collection_has_type_int(collection: Collection[NumberType]) -> bool:
    return collection_has_type(collection, int, False) or collection_has_type(collection, np.integer, True)


def collection_has_type_float(collection: Collection[NumberType]) -> bool:
    return collection_has_type(collection, float, False) or collection_has_type(collection, np.floating, True)


def is_iter(x: Any) -> bool:
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


def get_iterable_elements_datatype(
    it: Union[numpy.typing.NDArray["NumberT"], Sequence["NumberT"], "NumberT"]
) -> Type["NumberT"]:  # type: ignore[type-var]
    if isinstance(it, np.ndarray):
        item = cast("NumberT", it[0].item())
        return type(item)

    elif isinstance(it, IterableClass):
        if len(get_all_iterable_data_types(it)) > 1:
            raise ValueError("Multiple datatypes encountered in iterable object")
        item = next(iter(it))
        if isinstance(item, np.generic):
            return type(item.item())  # type: ignore[attr-defined]
        else:
            return type(item)
    else:
        raise ValueError(f"Did not found the type of {it}, maybe it's not an iterable")



import sys
import asyncio
import selectors
import threading
from typing import Any, TypeVar, Coroutine

from qm.utils.general_utils import Singleton


class EventLoopThread(metaclass=Singleton):
    def __init__(self) -> None:
        if sys.platform == "win32":
            loop = asyncio.SelectorEventLoop(selectors.SelectSelector())
        else:
            loop = asyncio.new_event_loop()
        self.loop = loop
        self._thread = threading.Thread(target=self.loop.run_forever)
        self._thread.daemon = True
        self._thread.start()

    def stop(self) -> None:
        self.loop.stop()
        self._thread.join()


T = TypeVar("T")


# TODO: in 3.9, remove the typing ignore


# In 3.9 future is generic: asyncio.Future[T]
def create_future(coroutine: Coroutine[Any, Any, T]) -> asyncio.Future:  # type: ignore[type-arg]
    return asyncio.run_coroutine_threadsafe(coroutine, loop=EventLoopThread().loop)  # type: ignore[return-value]


def run_async(coroutine: Coroutine[Any, Any, T]) -> T:
    return create_future(coroutine).result()  # type: ignore[no-any-return]



from octave_sdk import (
    IFMode,
    ClockType,
    OctaveOutput,
    RFOutputMode,
    ClockFrequency,
    OctaveLOSource,
    RFInputLOSource,
    RFInputRFSource,
)

from qm.octave.octave_manager import ClockMode
from qm.octave.calibration_db import CalibrationDB
from qm.octave.octave_config import QmOctaveConfig
from qm.octave.calibration_utils import convert_to_correction
from qm.octave.octave_mixer_calibration import MixerCalibrationResults
from qm.octave.abstract_calibration_db import AbstractCalibrationDB, AbstractIFCalibration, AbstractLOCalibration

__all__ = [
    "OctaveOutput",
    "ClockType",
    "ClockFrequency",
    "ClockMode",
    "OctaveLOSource",
    "IFMode",
    "RFInputLOSource",
    "RFInputRFSource",
    "RFOutputMode",
    "QmOctaveConfig",
    "CalibrationDB",
    "AbstractLOCalibration",
    "AbstractIFCalibration",
    "AbstractCalibrationDB",
    "convert_to_correction",
    "MixerCalibrationResults",
]



import logging
from typing import List, Type, Tuple, Optional, cast

from betterproto.lib.google.protobuf import Empty

import qm.grpc.qm_api
from qm.type_hinting import Value
from qm.grpc.qua import QuaProgram
from qm.grpc.compiler import QuaValues
from qm.utils.async_utils import run_async
from qm.grpc.general_messages import Matrix
from qm.api.models.jobs import InsertDirection
from qm.utils.protobuf_utils import LOG_LEVEL_MAP
from qm.api.models.capabilities import OPX_FEM_IDX
from qm.api.base_api import BaseApi, connection_error_handle
from qm.api.models.quantum_machine import QuantumMachineData
from qm.api.models.devices import Polarity, MixerInfo, AnalogOutputPortFilter
from qm.grpc.qua_config import QuaConfig, QuaConfigFemTypes, QuaConfigDeviceDec
from qm.api.models.compiler import CompilerOptionArguments, get_request_compiler_options
from qm.grpc.qm_manager import (
    Controller,
    GetQuantumMachineRequest,
    OpenQuantumMachineRequest,
    CloseQuantumMachineRequest,
)
from qm.grpc.frontend import (
    FrontendStub,
    QmDataRequest,
    QueuePosition,
    CompileRequest,
    IoValueRequest,
    AddToQueueRequest,
    ExecutionOverrides,
    AddCompiledToQueueRequest,
    ResetDataProcessingRequest,
    PerformHalDebugCommandRequest,
)
from qm.exceptions import (
    QMRequestError,
    OpenQmException,
    QMConnectionError,
    QMHealthCheckError,
    QMRequestDataError,
    CompilationException,
    FailedToAddJobToQueueException,
    QMFailedToGetQuantumMachineError,
    QmFailedToCloseQuantumMachineError,
    QMFailedToCloseAllQuantumMachinesError,
)
from qm.grpc.qm_api import (
    DigitalInputPort,
    HighQmApiRequest,
    HighQmApiRequestQePort,
    HighQmApiRequestSetIoValues,
    HighQmApiRequestSetFrequency,
    HighQmApiRequestSetCorrection,
    HighQmApiRequestIoValueSetData,
    HighQmApiRequestSetDigitalRoute,
    HighQmApiRequestSetInputDcOffset,
    HighQmApiRequestSetOutputDcOffset,
    HighQmApiRequestSetOutputFilterTaps,
    HighQmApiRequestAnalogOutputPortFilter,
    HighQmApiRequestSetCorrectionMixerInfo,
    HighQmApiRequestSetDigitalInputDeadtime,
    HighQmApiRequestSetDigitalInputPolarity,
    HighQmApiRequestSetDigitalInputThreshold,
)

logger = logging.getLogger(__name__)


@connection_error_handle()
class FrontendApi(BaseApi[FrontendStub]):
    @property
    def _stub_class(self) -> Type[FrontendStub]:
        return FrontendStub

    def get_version(self) -> str:
        response = run_async(self._stub.get_version(Empty(), timeout=self._timeout))
        return response.value

    def healthcheck(self, strict: bool) -> None:
        logger.info("Performing health check")
        response = run_async(self._stub.health_check(Empty(), timeout=self._timeout))

        for warning in response.warning_messages:
            logger.warning(f"Health check warning: {warning}")

        if not response.ok:
            logger.error(f"Health check error: {response.message}")

            for error in response.error_messages:
                logger.error(f"Health check error: {error}")

            if strict:
                raise QMHealthCheckError("Health check failed")
            return

        logger.info("Health check passed")

    def reset_data_processing(self) -> None:
        run_async(self._stub.reset_data_processing(ResetDataProcessingRequest(), timeout=self._timeout))

    def open_qm(self, config: QuaConfig, close_other_machines: bool, keep_dc_offsets_when_closing: bool = False) -> str:
        request = OpenQuantumMachineRequest(config=config, keep_dc_offsets_when_closing=keep_dc_offsets_when_closing)

        if close_other_machines:
            request.always = True
        else:
            request.never = True

        response = run_async(self._stub.open_quantum_machine(request, timeout=self._timeout))

        if not response.success:
            error_messages = []
            for error in response.config_validation_errors:
                error_messages.append(f'CONFIG ERROR in key "{error.path}" [{error.group}] : {error.message}')

            for physical_error in response.physical_validation_errors:
                error_messages.append(
                    f"PHYSICAL CONFIG ERROR in key "
                    f'"{physical_error.path}" [{physical_error.group}] : {physical_error.message}'
                )

            for msg in error_messages:
                logger.error(msg)

            error_details = [(item.group, item.path, item.message) for item in response.config_validation_errors] + [
                (item.group, item.path, item.message) for item in response.physical_validation_errors
            ]
            formatted_errors = "\n".join(error_messages)
            raise OpenQmException(
                f"Can not open QM, see the following errors:\n{formatted_errors}", errors=error_details
            )

        for warning in response.open_qm_warnings:
            logger.warning(f"Open QM ended with warning {warning.code}: {warning.message}")

        return response.machine_id

    def list_open_quantum_machines(self) -> List[str]:
        response = run_async(self._stub.list_open_quantum_machines(Empty(), timeout=self._timeout))
        return response.machine_i_ds

    def get_quantum_machine(self, qm_id: str) -> QuantumMachineData:
        request = GetQuantumMachineRequest(machine_id=qm_id)
        response = run_async(self._stub.get_quantum_machine(request, timeout=self._timeout))

        if not response.success:
            error_message = "\n".join([error.message for error in response.errors])
            raise QMFailedToGetQuantumMachineError(f"Failed to fetch quantum machine: {error_message}")

        return QuantumMachineData(machine_id=response.machine_id, config=response.config)

    def close_quantum_machine(self, machine_id: str) -> bool:
        request = CloseQuantumMachineRequest(machine_id=machine_id)
        response = run_async(self._stub.close_quantum_machine(request, timeout=self._timeout))
        if not response.success:
            raise QmFailedToCloseQuantumMachineError("\n".join(err.message for err in response.errors))
        return True

    def get_quantum_machine_config(self, machine_id: str) -> QuaConfig:
        machine_data = self.get_quantum_machine(machine_id)
        config = machine_data.config
        # TODO - this is a patch to make the tests work, once we move to the new GRPC messages,
        #  we will need to check it for backwards compatibility.
        if not config.v1_beta.control_devices:
            for controller_name, controller in config.v1_beta.controllers.items():
                config.v1_beta.control_devices[controller_name] = QuaConfigDeviceDec(
                    fems={OPX_FEM_IDX: QuaConfigFemTypes(opx=controller)}
                )
        return machine_data.config

    def close_all_quantum_machines(self) -> None:
        response = run_async(self._stub.close_all_quantum_machines(Empty(), timeout=self._timeout))
        if not response.success:
            messages = [error.message for error in response.errors]
            for msg in messages:
                logger.error(msg)

            raise QMFailedToCloseAllQuantumMachinesError(
                "Can not close all quantum machines. See the following errors:\n" + "\n".join(messages),
            )

    def get_controllers(self) -> List[Controller]:
        response = run_async(self._stub.get_controllers(Empty(), timeout=self._timeout))
        return response.controllers

    def clear_all_job_results(self) -> None:
        run_async(self._stub.clear_all_job_results(Empty(), timeout=self._timeout))

    def send_debug_command(self, controller_name: str, command: str) -> str:
        request = PerformHalDebugCommandRequest(controller_name=controller_name, command=command)
        response = run_async(self._stub.perform_hal_debug_command(request, timeout=self._timeout))

        if not response.success:
            raise QMConnectionError(response.response)
        return response.response

    def add_to_queue(
        self,
        machine_id: str,
        program: QuaProgram,
        compiler_options: CompilerOptionArguments,
        insert_direction: InsertDirection,
    ) -> str:
        queue_position = QueuePosition()
        if insert_direction == InsertDirection.start:
            queue_position.start = Empty()
        elif insert_direction == InsertDirection.end:
            queue_position.end = Empty()

        program.compiler_options = get_request_compiler_options(compiler_options)

        request = AddToQueueRequest(
            quantum_machine_id=machine_id,
            high_level_program=program,
            queue_position=queue_position,
        )

        logger.info("Sending program to QOP for compilation")

        response = run_async(self._stub.add_to_queue(request, timeout=None))

        for message in response.messages:
            logger.log(LOG_LEVEL_MAP[message.level], message.message)

        job_id = response.job_id
        if not response.ok:
            logger.error(f"Job {job_id} failed. Failed to execute program.")
            raise FailedToAddJobToQueueException(f"Job {job_id} failed. Failed to execute program.")

        return job_id

    def add_compiled_to_queue(self, machine_id: str, program_id: str, execution_overrides: ExecutionOverrides) -> str:
        queue_position = QueuePosition()
        queue_position.end = Empty()

        request = AddCompiledToQueueRequest(
            quantum_machine_id=machine_id,
            program_id=program_id,
            queue_position=queue_position,
            execution_overrides=execution_overrides,
        )

        response = run_async(self._stub.add_compiled_to_queue(request, timeout=self._timeout))

        job_id = response.job_id

        for err in response.errors:
            logger.error(err.message)

        if not response.ok:
            logger.error(f"Job {job_id} failed. Failed to execute program.")
            raise FailedToAddJobToQueueException(f"Job {job_id} failed. Failed to execute program.")

        return job_id

    def compile(
        self,
        machine_id: str,
        program: QuaProgram,
        compiler_options: CompilerOptionArguments,
    ) -> str:
        program.compiler_options = get_request_compiler_options(compiler_options)
        request = CompileRequest(quantum_machine_id=machine_id, high_level_program=program)

        response = run_async(self._stub.compile(request, timeout=None))

        for message in response.messages:
            logger.log(LOG_LEVEL_MAP[message.level], message.message)

        program_id = response.program_id
        if not response.ok:
            logger.error(f"Compilation of program {program_id} failed")
            raise CompilationException(f"Compilation of program {program_id} failed")
        return program_id

    def _perform_qm_request(self, request: HighQmApiRequest) -> None:
        response = run_async(self._stub.perform_qm_request(request, timeout=self._timeout))

        if not response.ok:
            error_message = "\n".join(it.message for it in response.errors)
            logger.error(f"Failed: {error_message}")
            raise QMRequestError(f"Failed: {error_message}")

    def set_correction(self, machine_id: str, mixer: MixerInfo, correction: Matrix) -> None:
        correction_request = HighQmApiRequestSetCorrection(
            mixer=HighQmApiRequestSetCorrectionMixerInfo(
                mixer=mixer.mixer,
                frequency_negative=mixer.frequency_negative,
                intermediate_frequency=mixer.intermediate_frequency,
                intermediate_frequency_double=mixer.intermediate_frequency_double,
                lo_frequency=mixer.lo_frequency,
                lo_frequency_double=mixer.lo_frequency_double,
            ),
            correction=cast(qm.grpc.qm_api.Matrix, correction),  # For some reason we have two matrix messages,
            # although they are the same...
        )
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_correction=correction_request)
        self._perform_qm_request(request)

    def set_intermediate_frequency(self, machine_id: str, element: str, value: float) -> None:
        set_frequency_request = HighQmApiRequestSetFrequency(qe=element, value=value)
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_frequency=set_frequency_request)
        self._perform_qm_request(request)

    def set_output_dc_offset(self, machine_id: str, element: str, element_port: str, offset: float) -> None:
        output_dc_offset_request = HighQmApiRequestSetOutputDcOffset(
            qe=HighQmApiRequestQePort(qe=element, port=element_port), i=offset, q=offset
        )
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_output_dc_offset=output_dc_offset_request)
        self._perform_qm_request(request)

    def set_output_filter_taps(
        self,
        machine_id: str,
        element: str,
        element_port: str,
        filter_port: AnalogOutputPortFilter,
    ) -> None:
        output_filter_tap_request = HighQmApiRequestSetOutputFilterTaps(
            qe=HighQmApiRequestQePort(qe=element, port=element_port),
            filter=HighQmApiRequestAnalogOutputPortFilter(
                feedback=filter_port.feedback, feedforward=filter_port.feedforward
            ),
        )
        request = HighQmApiRequest(
            quantum_machine_id=machine_id,
            set_output_filter_taps=output_filter_tap_request,
        )
        self._perform_qm_request(request)

    def set_input_dc_offset(self, machine_id: str, element: str, element_port: str, offset: float) -> None:
        input_dc_offset_request = HighQmApiRequestSetInputDcOffset(
            qe=HighQmApiRequestQePort(qe=element, port=element_port), offset=offset
        )
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_input_dc_offset=input_dc_offset_request)
        self._perform_qm_request(request)

    def set_output_digital_delay(self, machine_id: str, element: str, element_port: str, delay: int) -> None:
        digital_delay_request = HighQmApiRequestSetDigitalRoute(
            delay=HighQmApiRequestQePort(qe=element, port=element_port), value=delay
        )
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_digital_route=digital_delay_request)
        self._perform_qm_request(request)

    def set_output_digital_buffer(self, machine_id: str, element: str, element_port: str, buffer: int) -> None:
        digital_buffer_request = HighQmApiRequestSetDigitalRoute(
            buffer=HighQmApiRequestQePort(qe=element, port=element_port), value=buffer
        )
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_digital_route=digital_buffer_request)
        self._perform_qm_request(request)

    def set_io_values(self, machine_id: str, values: List[Optional[Value]]) -> None:
        type_to_value_mapping = {
            int: "int_value",
            float: "double_value",
            bool: "boolean_value",
        }

        set_data = []
        for index, value in enumerate(values):
            if value is not None:
                set_data_request = HighQmApiRequestIoValueSetData(io_number=index + 1)
                setattr(set_data_request, type_to_value_mapping[type(value)], value)
                set_data.append(set_data_request)

        set_io_values = HighQmApiRequestSetIoValues(all=True, io_value_set_data=set_data)
        request = HighQmApiRequest(quantum_machine_id=machine_id, set_io_values=set_io_values)
        self._perform_qm_request(request)

    def set_digital_input_threshold(
        self, machine_id: str, controller_name: str, fem_number: int, port_number: int, threshold: float
    ) -> None:
        digital_input_threshold_request = HighQmApiRequestSetDigitalInputThreshold(
            digital_port=DigitalInputPort(
                controller_name=controller_name, fem_number=fem_number, port_number=port_number
            ),
            threshold=threshold,
        )
        request = HighQmApiRequest(
            quantum_machine_id=machine_id,
            set_digital_input_threshold=digital_input_threshold_request,
        )
        self._perform_qm_request(request)

    def set_digital_input_dead_time(
        self, machine_id: str, controller_name: str, fem_number: int, port_number: int, dead_time: int
    ) -> None:
        digital_input_dead_time_request = HighQmApiRequestSetDigitalInputDeadtime(
            digital_port=DigitalInputPort(
                controller_name=controller_name, fem_number=fem_number, port_number=port_number
            ),
            deadtime=dead_time,
        )
        request = HighQmApiRequest(
            quantum_machine_id=machine_id,
            set_digital_input_deadtime=digital_input_dead_time_request,
        )
        self._perform_qm_request(request)

    def set_digital_input_polarity(
        self,
        machine_id: str,
        controller_name: str,
        fem_number: int,
        port_number: int,
        polarity: Polarity,
    ) -> None:
        digital_input_polarity_request = HighQmApiRequestSetDigitalInputPolarity(
            digital_port=DigitalInputPort(
                controller_name=controller_name, fem_number=fem_number, port_number=port_number
            ),
            polarity=polarity.value,
        )
        request = HighQmApiRequest(
            quantum_machine_id=machine_id,
            set_digital_input_polarity=digital_input_polarity_request,
        )
        self._perform_qm_request(request)

    def get_io_values(self, machine_id: str) -> Tuple[QuaValues, ...]:
        request = QmDataRequest(
            io_value_request=[
                IoValueRequest(io_number=1, quantum_machine_id=machine_id),
                IoValueRequest(io_number=2, quantum_machine_id=machine_id),
            ]
        )
        response = run_async(self._stub.request_data(request, timeout=self._timeout))

        if not response.success:
            raise QMRequestDataError("\n".join(err.message for err in response.errors))

        return tuple(resp.values for resp in response.io_value_response)



import json
import logging
import warnings
from typing import List, Tuple, Union, Literal, Optional, Sequence, overload

from qm.api.v2.job_api import JobApi
from qm.octave import QmOctaveConfig
from qm.persistence import BaseStore
from qm.program.program import Program
from qm.utils import deprecation_message
from qm.exceptions import FunctionInputError
from qm.octave.octave_manager import OctaveManager
from qm.simulate.interface import SimulationConfig
from qm.type_hinting.config_types import DictQuaConfig
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.qm_api import QmApi, IoValue, NoRunningQmJob
from qm.api.v2.job_api.job_api import JobApiWithDeprecations
from qm.type_hinting import Value, Number, NumpySupportedValue
from qm.type_hinting.general import PathLike, NumpySupportedFloat
from qm.jobs.job_queue_with_deprecations import QmQueueWithDeprecations
from qm.utils.config_utils import get_fem_config, element_has_mix_inputs
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApiWithDeprecations
from qm.api.models.compiler import CompilerOptionArguments, standardize_compiler_params
from qm.grpc.qua_config import QuaConfig, QuaConfigMicrowaveFemDec, QuaConfigAdcPortReference, QuaConfigDacPortReference

logger = logging.getLogger(__name__)


class QmApiWithDeprecations(QmApi):
    SIMULATED_JOB_CLASS = SimulatedJobApiWithDeprecations

    def __init__(
        self,
        connection_details: ConnectionDetails,
        qm_id: str,
        store: BaseStore,
        capabilities: ServerCapabilities,
        octave_config: Optional[QmOctaveConfig],
        octave_manager: OctaveManager,
        pb_config: Optional[QuaConfig] = None,
    ):
        super().__init__(connection_details, qm_id, store, capabilities, octave_config, octave_manager, pb_config)
        self._queue = QmQueueWithDeprecations(store=store, api=self, capabilities=self._caps)

    def _get_job(self, job_id: str, store: BaseStore) -> JobApiWithDeprecations:
        return JobApiWithDeprecations(self.connection_details, job_id, store, capabilities=self._caps)

    def get_job_by_id(self, job_id: str) -> JobApiWithDeprecations:
        """
        Deprecated - This method is going to be removed, please use `qmm.get_job()`.

        Returns the job object for the given id.
        Args:
            job_id: The job id
        Returns:
             The job object
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_job_by_id",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please use `qmm.get_job()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._get_job(job_id, self._store)

    @property
    def queue(self) -> QmQueueWithDeprecations:
        warnings.warn(
            deprecation_message(
                method="qm.queue",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This property is going to be removed, all functionality will exist directly under "
                "`QuantumMachine`. For example, instead of `qm.queue.add(prog)` use `qm.add_to_queue(prog)`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._queue

    def close(self) -> bool:  # type: ignore[override]
        """Closes the quantum machine.

        Returns:
             `True` if the close request succeeded.
        """
        super().close()
        return True

    def execute(
        self,
        program: Program,
        duration_limit: None = None,
        data_limit: None = None,
        force_execution: None = None,
        dry_run: None = None,
        simulate: Optional[SimulationConfig] = None,
        *,
        compiler_options: Optional[CompilerOptionArguments] = None,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> JobApi:
        """Executes a program and returns a job object to keep track of execution and get
        results.

        Note:

            Calling execute will halt any currently running program and clear the current
            queue. If you want to add a job to the queue, use qm.queue.add()

        Args:
            program: A QUA ``program()`` object to execute
            duration_limit: This parameter is ignored and will be removed in future versions
            data_limit: This parameter is ignored and will be removed in future versions
            force_execution: This parameter is ignored and will be removed in future versions
            dry_run: This parameter is ignored and will be removed in future versions
            simulate: If given, will be simulated instead of executed.
            compiler_options: Optional arguments for compilation.
            strict: This parameter is deprecated, please use `compiler_options`
            flags: This parameter is deprecated, please use `compiler_options`
        Returns:
            A ``QmJob`` object (see QM Job API).
        """
        if not isinstance(program, Program):
            raise Exception("program argument must be of type qm.program.Program")

        for x, name in [
            (duration_limit, "`duration_limit'"),
            (data_limit, "`data_limit'"),
            (force_execution, "`force_execution'"),
            (dry_run, "`dry_run'"),
        ]:
            if x is not None:
                warnings.warn(
                    deprecation_message(
                        method=f"The argument {name}",
                        deprecated_in="1.2.0",
                        removed_in="1.4.0",
                    ),
                )

        compiler_options = standardize_compiler_params(compiler_options, strict, flags)

        if simulate is not None:
            warnings.warn(
                deprecation_message(
                    method="The argument simulate",
                    deprecated_in="1.2.0",
                    removed_in="1.4.0",
                    details="The simulate argument is deprecated, please use the simulate method.",
                ),
            )
            return self.simulate(program, simulate, compiler_options=compiler_options)

        logger.info("Clearing queue")
        self.clear_queue()
        current_running_job = self._get_running_job()
        if current_running_job is not None:
            logger.info(f"Cancelling currently running job - {current_running_job.id}")
            current_running_job.cancel()

        new_job_api = self.add_to_queue(program, compiler_options=compiler_options)
        new_job_api.wait_until({"Running"}, timeout=5 * 60)
        # The timeout here is just for the backwards compatibility behaviour.
        # See that in the father function there is no `wait_until`
        return new_job_api

    def list_controllers(self) -> Tuple[str, ...]:
        """
        Deprecated - This method is going to be removed, please use `qm.get_config()`.

        Gets a list with the defined controllers in this qm

        Returns:
            The names of the controllers configured in this qm
        """
        warnings.warn(
            deprecation_message(
                method="qm.list_controllers",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please get data from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        return tuple(config.v1_beta.control_devices) or tuple(config.v1_beta.controllers)

    def set_mixer_correction(
        self,
        mixer: str,
        intermediate_frequency: Number,
        lo_frequency: Number,
        values: Tuple[float, float, float, float],
    ) -> None:
        """Deprecated - This method is going to be removed, please use `job.set_element_correction()`.

        Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer for the supplied intermediate frequency and LO frequency.

        Args:
            mixer (str): the name of the mixer, as defined in the
                configuration
            intermediate_frequency (Union[int|float]): the intermediate
                frequency for which to apply the correction matrix
            lo_frequency (int): the LO frequency for which to apply the
                correction matrix
            values (tuple):

                tuple is of the form (v00, v01, v10, v11) where
                the matrix is
                | v00 v01 |
                | v10 v11 |

        Note:

            Currently, the OPX does not support multiple mixer calibration entries.
            This function will accept IF & LO frequencies written in the config file,
            and will update the correction matrix for all the elements with the given
            mixer/frequencies combination when the program started.

            It’s not recommended to use this method while a job is running.
            To change the calibration values for a running job,
            use job.set_element_correction
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_mixer_correction",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please use `job.set_element_correction()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config: DictQuaConfig = {
            "version": 1,
            "mixers": {
                mixer: [
                    {
                        "correction": values,
                        "intermediate_frequency": intermediate_frequency,
                        "lo_frequency": lo_frequency,
                    }
                ]
            },
        }
        self.update_config(config)
        job = self._get_running_job()
        if job is not None:
            pb_config = self._get_pb_config()
            for name, element_config in pb_config.v1_beta.elements.items():
                if not element_has_mix_inputs(element_config):
                    continue
                mixer_cond = element_config.mix_inputs.mixer == mixer
                if_cond = element_config.intermediate_frequency == intermediate_frequency
                lo_cond = element_config.mix_inputs.lo_frequency == lo_frequency
                if mixer_cond and if_cond and lo_cond:
                    job.set_element_correction(name, values)

    def set_intermediate_frequency(self, element: str, freq: float) -> None:
        """
        Deprecated - This method is going to be moved to the job API, please use `job.set_intermediate_frequency()`.

        Sets the intermediate frequency of the element

        Args:
            element (str): the name of the element whose intermediate
                frequency will be updated
            freq (float): the intermediate frequency to set to the given
                element
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_intermediate_frequency",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.set_intermediate_frequency()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        job.set_intermediate_frequency(element, freq)

    def get_intermediate_frequency(self, element: str) -> float:
        """
        Deprecated - This method is going to be moved to the job API, please use `job.get_intermediate_frequency()`.

        Gets the intermediate frequency of the element

        Args:
            element (str): the name of the element whose intermediate
                frequency will be updated

        Returns:
            The intermediate frequency
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_intermediate_frequency",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.get_intermediate_frequency()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_intermediate_frequency(element)

    def get_output_dc_offset_by_element(
        self, element: str, iq_input: Optional[Literal["I", "Q", "single"]] = None
    ) -> float:
        """Deprecated - This method is going to be removed, please get idle value from `qm.get_config()`
            or current value from job `job.get_output_dc_offset_by_element()`

        Get the current DC offset of the OPX analog output channel associated with an element.

        Args:
            element: the name of the element to get the correction for
            iq_input: the port name as appears in the element config.
                Options:

                `'single'`
                    for an element with a single input

                `'I'` or `'Q'`
                    for an element with mixer inputs

        Returns:
            the offset, in volts
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_output_dc_offset_by_element",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please get idle value from `qm.get_config()`"
                " or current value from job `job.get_output_dc_offset_by_element()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_output_dc_offset_by_element(element, iq_input)

    @overload
    def set_output_dc_offset_by_element(self, element: str, input: Literal["single", "I", "Q"], offset: float) -> None:
        pass

    @overload
    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Tuple[Literal["I", "Q"], Literal["I", "Q"]],
        offset: Union[Tuple[float, float], List[float]],
    ) -> None:
        pass

    def set_output_dc_offset_by_element(
        self,
        element: str,
        input: Union[Literal["single", "I", "Q"], Tuple[Literal["I", "Q"], Literal["I", "Q"]], List[Literal["I", "Q"]]],
        offset: Union[float, Tuple[float, float], List[float]],
    ) -> None:
        """Deprecated - This method is going to be removed, please set idle value with `qm.update_config()` or current
            value from job `job.set_output_dc_offset_by_element()`

        Set the current DC offset of the OPX analog output channel associated with an element.

        Args:
            element (str): the name of the element to update the
                correction for
            input (Union[str, Tuple[str,str], List[str]]): the input
                name as appears in the element config. Options:

                `'single'`
                    for an element with a single input

                `'I'` or `'Q'` or a tuple ('I', 'Q')
                    for an element with mixer inputs
            offset (Union[float, Tuple[float,float], List[float]]): The
                dc value to set to, in volts.

        Examples:
            ```python
            qm.set_output_dc_offset_by_element('flux', 'single', 0.1)
            qm.set_output_dc_offset_by_element('qubit', 'I', -0.01)
            qm.set_output_dc_offset_by_element('qubit', ('I', 'Q'), (-0.01, 0.05))
            ```

        Note:

            If the sum of the DC offset and the largest waveform data-point exceed the range,
            DAC output overflow will occur and the output will be corrupted.
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_output_dc_offset_by_element",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please set idle value with `qm.update_config()`"
                " or current value from job `job.set_output_dc_offset_by_element()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        if isinstance(input, str):
            if not isinstance(offset, (int, float)):
                raise FunctionInputError(f"Input should be int or float, got {type(offset)}")
            if input in {"I", "Q"}:
                ports = self._get_input_ports_from_mixed_input_element(element)
                port = ports[0] if input == "I" else ports[1]
            else:
                port = self._get_input_port_from_single_input_element(element)
            config = self._create_config_for_output_dc_offset_setting(port, offset)
            self.update_config(config)
            job = self._get_running_job()
            if job is not None:
                job.set_output_dc_offset_by_element(element, input, offset)
        elif isinstance(input, (list, tuple)):
            if not set(input) <= {"I", "Q"}:
                raise FunctionInputError(f"Input names should be 'I' or 'Q', got {input}")
            if not (isinstance(offset, (list, tuple)) and len(input) == len(offset)):
                raise FunctionInputError(
                    f"input should be two iterables of the same size," f"got input = {input} and offset = {offset}"
                )
            ports = self._get_input_ports_from_mixed_input_element(element)
            for _input, _offset in zip(input, offset):
                _port = ports[0] if _input == "I" else ports[1]
                config = self._create_config_for_output_dc_offset_setting(_port, _offset)
                self.update_config(config)
            job = self._get_running_job()
            if job is not None:
                job.set_output_dc_offset_by_element(element, input, offset)
        else:
            raise FunctionInputError(f"Input should be str or tuple, got {type(input)}")

    def set_output_filter_by_element(
        self,
        element: str,
        input: str,
        feedforward: Sequence[NumpySupportedFloat],
        feedback: Sequence[NumpySupportedFloat],
    ) -> None:
        raise NotImplementedError

    def set_input_dc_offset_by_element(self, element: str, output: str, offset: float) -> None:
        """Deprecated - This method is going to be moved to the job API, please use job.set_input_dc_offset_by_element()`

        Set the current DC offset of the OPX analog input channel associated with an element.

        Args:
            element (str): the name of the element to update the
                correction for
            output (str): the output key name as appears in the element
                config under 'outputs'.
            offset (float): the dc value to set to, in volts. Ranges from -0.5 to 0.5 - 2^-16 in steps of
                2^-16.

        Note:
            If the sum of the DC offset and the largest waveform data-point exceed the range,
            DAC output overflow will occur and the output will be corrupted.
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_input_dc_offset_by_element",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.set_input_dc_offset_by_element()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        port = self._get_output_port_from_element(element, output)
        config = self._create_config_for_input_dc_offset_setting(port, offset)
        self.update_config(config)

        job = self._get_running_job()
        if job is not None:
            job.set_input_dc_offset_by_element(element, output, offset)

    def get_input_dc_offset_by_element(self, element: str, output: str) -> float:
        """Deprecated - This method is going to be removed, please get the value from `qm.get_config()`.

        Get the current DC offset of the OPX analog input channel associated with an element.

        Args:
            element: the name of the element to get the correction for
            output: the output key name as appears in the element config
                under 'outputs'.

        Returns:
            The offset, in volts
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_input_dc_offset_by_element",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please get the value from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        port = self._get_output_port_from_element(element, output)
        fem_config = get_fem_config(config, port)
        if isinstance(fem_config, QuaConfigMicrowaveFemDec):
            raise ValueError(f"Element {element} does not support dc offset.")
        return fem_config.analog_inputs[port.number].offset

    def get_digital_delay(self, element: str, digital_input: str) -> int:
        """Deprecated - This method is going to be moved to the job API, please use `job.get_output_digital_delay()`.

        Gets the delay of the digital input of the element

        Args:
            element: The name of the element to get the delay for
            digital_input: The digital input name as appears in the
                element's config

        Returns:
            The delay
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_digital_delay",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.get_output_digital_delay()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_output_digital_buffer(element, digital_input)

    def set_digital_delay(self, element: str, digital_input: str, delay: int) -> None:
        """Deprecated - This method is going to be moved to the job API, please use `job.set_output_digital_delay()`.

        Sets the delay of the digital input of the element

        Args:
            element (str): The name of the element to update delay for
            digital_input (str): The digital input name as appears in
                the element's config
            delay (int): The delay value to set to, in ns.
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_digital_delay",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.set_output_digital_delay()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        job.set_output_digital_delay(element, digital_input, delay)

    def get_digital_buffer(self, element: str, digital_input: str) -> int:
        """Deprecated - This method is going to be moved to the job API, please use `job.get_output_digital_buffer()`.

        Gets the buffer for digital input of the element

        Args:
            element (str): The name of the element to get the buffer for
            digital_input (str): The digital input name as appears in
                the element's config

        Returns:
            The buffer
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_digital_buffer",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.get_output_digital_buffer()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        return job.get_output_digital_buffer(element, digital_input)

    def set_digital_buffer(self, element: str, digital_input: str, buffer: int) -> None:
        """Deprecated - This method is going to be moved to the job API, please use `job.set_output_digital_buffer()`.

        Sets the buffer for digital input of the element

        Args:
            element (str): The name of the element to update buffer for
            digital_input (str): the digital input name as appears in
                the element's config
            buffer (int): The buffer value to set to, in ns.
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_digital_buffer",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use "
                "`job.set_output_digital_buffer()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        job = self._strict_get_running_job()
        job.set_output_digital_buffer(element, digital_input, buffer)

    def get_time_of_flight(self, element: str) -> int:
        """Deprecated - This method is going to be removed, please get the value from `qm.get_config()`.

        Gets the *time of flight*, associated with a measurement element.

        This is the amount of time between the beginning of a measurement pulse applied to element
        and the time that the data is available to the controller for demodulation or streaming.

        Args:
            element (str): The name of the element to get time of flight
                for

        Returns:
            The time of flight, in ns
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_time_of_flight",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please get the value from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        tof = config.v1_beta.elements[element].time_of_flight
        if tof is None:
            raise ValueError(f"Time of flight for element {element} is not set")
        return tof

    def get_smearing(self, element: str) -> int:
        """Deprecated - This method is going to be removed, please get the value from `qm.get_config()`.

        Gets the *smearing* associated with a measurement element.

        This is a broadening of the raw results acquisition window, to account for dispersive broadening
        in the measurement elements (readout resonators etc.) The acquisition window will be broadened
        by this amount on both sides.

        Args:
            element (str): The name of the element to get smearing for

        Returns:
            The smearing, in ns.
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_smearing",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please get the value from `qm.get_config()`.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        config = self._get_pb_config()
        smearing = config.v1_beta.elements[element].smearing
        if smearing is None:
            raise ValueError(f"Smearing for element {element} is not set")
        return smearing

    @property
    def io1(self) -> IoValue:
        warnings.warn(
            deprecation_message(
                method="qm.io1",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This property is going to be removed, please use `job.get_io_values()[0]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io1_value()

    @io1.setter
    def io1(self, value: Value) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.io1",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This property is going to be removed, please use `job.set_io_values(io1=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io1_value(value)

    @property
    def io2(self) -> IoValue:
        warnings.warn(
            deprecation_message(
                method="qm.io2",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This property is going to be removed, please use `job.get_io_values()[1]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io1_value()

    @io2.setter
    def io2(self, value: Value) -> None:
        warnings.warn(
            deprecation_message(
                method="qm.io2",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This property is going to be removed, please use `job.set_io_values(io2=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io2_value(value)

    def set_io1_value(self, value_1: Value) -> None:
        """Deprecated - This method is going to be moved to the job API, please use `job.set_io_values(io1=value)`

        Sets the values of ``IO1``

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``, ``value_2``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value_1: The value to be placed in ``IO1``
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_io1_value",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use `job.set_io_values(io1=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io_values(value_1=value_1)

    def set_io2_value(self, value_2: Value) -> None:
        """Deprecated - This method is going to be moved to the job API, please use `job.set_io_values(io2=value)`

        Sets the values of ``IO2``

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``, ``value_2``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value_2: The value to be placed in ``IO2``
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_io2_value",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use `job.set_io_values(io2=value)`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        self.set_io_values(value_2=value_2)

    def set_io_values(
        self,
        value_1: Optional[NumpySupportedValue] = None,
        value_2: Optional[NumpySupportedValue] = None,
    ) -> None:
        """Deprecated - This method is going to be moved to the job API, please use `job.set_io_values()`

        Sets the values of ``IO1`` & ``IO2`

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``, ``value_2``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        Args:
            value_1: The value to be placed in ``IO1``
            value_2: The value to be placed in ``IO2``
        """
        warnings.warn(
            deprecation_message(
                method="qm.set_io_values",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use `job.set_io_values()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        if value_1 is None and value_2 is None:
            return

        job = self._strict_get_running_job()
        job.set_io_values(value_1, value_2)

    def get_io1_value(self) -> IoValue:
        """Deprecated - This method is going to be moved to the job API, please use `job.get_io_values()[0]`

        Gets the data stored in ``IO1``

        No inference is made on type.

        Returns:
            A dictionary with data stored in ``IO1``. (Data is in all
            three format: ``int``, ``float`` and ``bool``)
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_io2_value",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use `job.get_io_values()[0]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io_values()[0]

    def get_io2_value(self) -> IoValue:
        """Deprecated - This method is going to be moved to the job API, please use `job.get_io_values()[1]`

        Gets the data stored in ``IO2``

        No inference is made on type.

        Returns:
            A dictionary with data stored in ``IO2``. (Data is in all
            three format: ``int``, ``float`` and ``bool``)
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_io2_value",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use `job.get_io_values()[1]`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self.get_io_values()[1]

    def get_io_values(self) -> List[IoValue]:
        """Deprecated - This method is going to be moved to the job API, please use `job.get_io_values()`

        Gets the data stored in ``IO1`` & ``IO2``

        No inference is made on type.

        Returns:
            A dictionary with data stored in ``IO1`` & ``IO2`` (Data is in all
            three format: ``int``, ``float`` and ``bool``)
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_io_values",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be moved to the job API, please use `job.get_io_values()`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        running_job = self._strict_get_running_job()
        resp1, resp2 = running_job.get_io_values()
        return [
            {
                "io_number": 1,
                "int_value": resp1.int_value,
                "fixed_value": resp1.double_value,
                "boolean_value": resp1.boolean_value,
            },
            {
                "io_number": 2,
                "int_value": resp2.int_value,
                "fixed_value": resp2.double_value,
                "boolean_value": resp2.boolean_value,
            },
        ]

    def save_config_to_file(self, filename: PathLike) -> None:
        """Deprecated - This method is going to be removed.

        Saves the qm current config to a file

        Args:
            filename: The name of the file where the config will be saved
        """
        warnings.warn(
            deprecation_message(
                method="qm.save_config_to_file",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        with open(filename, "w") as writer:
            json.dump(self.get_config(), writer)

    def get_running_job(self) -> Optional[JobApiWithDeprecations]:
        """Deprecated - This method is going to be removed, please use `qm.get_jobs(status=['Running'])`

        Gets the currently running job. Returns None if there isn't one.
        """
        warnings.warn(
            deprecation_message(
                method="qm.get_running_job",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method is going to be removed, please use `qm.get_jobs(status=['Running'])`",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        return self._get_running_job()

    def _get_running_job(self) -> Optional[JobApiWithDeprecations]:
        jobs = self.get_jobs(status=["Running"])
        if jobs:
            return self._get_job(jobs[0].id, store=self._store)
        return None

    def _strict_get_running_job(self) -> JobApiWithDeprecations:
        job = self._get_running_job()
        if job is None:
            raise NoRunningQmJob("No running job found")
        return job

    def _get_output_port_from_element(self, element_name: str, port_name: str) -> QuaConfigAdcPortReference:
        element = self._get_pb_config().v1_beta.elements[element_name]
        return element.multiple_outputs.port_references[port_name]

    def _get_input_ports_from_mixed_input_element(
        self, element_name: str
    ) -> Tuple[QuaConfigDacPortReference, QuaConfigDacPortReference]:
        element = self._get_pb_config().v1_beta.elements[element_name]
        return element.mix_inputs.i, element.mix_inputs.q

    def _get_input_port_from_single_input_element(self, element_name: str) -> QuaConfigDacPortReference:
        element = self._get_pb_config().v1_beta.elements[element_name]
        return element.single_input.port

    @staticmethod
    def _create_config_for_input_dc_offset_setting(port: QuaConfigAdcPortReference, value: Number) -> DictQuaConfig:
        return {
            "version": 1,
            "controllers": {
                port.controller: {"fems": {port.fem: {"type": "LF", "analog_inputs": {port.number: {"offset": value}}}}}
            },
        }

    @staticmethod
    def _create_config_for_output_dc_offset_setting(port: QuaConfigDacPortReference, value: Number) -> DictQuaConfig:
        return {
            "version": 1,
            "controllers": {
                port.controller: {
                    "fems": {port.fem: {"type": "LF", "analog_outputs": {port.number: {"offset": value}}}}
                }
            },
        }



import datetime
import warnings
from typing import List, Optional

import betterproto

from qm.type_hinting import Value
from qm.persistence import BaseStore
from qm.exceptions import QmQuaException
from qm.utils import deprecation_message
from qm.api.frontend_api import FrontendApi
from qm.api.models.capabilities import ServerCapabilities
from qm.api.job_manager_api import create_job_manager_from_api
from qm.grpc.frontend import (
    JobExecutionStatus,
    JobExecutionStatusLoading,
    JobExecutionStatusPending,
    JobExecutionStatusRunning,
    JobExecutionStatusCompleted,
)


class QmBaseJob:
    def __init__(
        self,
        job_id: str,
        machine_id: str,
        frontend_api: FrontendApi,
        capabilities: ServerCapabilities,
        store: BaseStore,
    ):
        self._id = job_id
        self._machine_id = machine_id
        self._frontend = frontend_api
        self._capabilities = capabilities
        self._store = store

        self._job_manager = create_job_manager_from_api(frontend_api)

        self._added_user_id: Optional[str] = None
        self._time_added: Optional[datetime.datetime] = None
        self._initialize_from_job_status()

    def _initialize_from_job_status(self) -> None:
        status: JobExecutionStatus = self._job_manager.get_job_execution_status(self._id, self._machine_id)
        _, job_state = betterproto.which_one_of(status, "status")

        if isinstance(
            job_state,
            (
                JobExecutionStatusPending,
                JobExecutionStatusRunning,
                JobExecutionStatusCompleted,
                JobExecutionStatusLoading,
            ),
        ):
            self._added_user_id = job_state.added_by
            self._time_added = job_state.time_added

    @property
    def status(self) -> str:
        """Returns the status of the job, one of the following strings:
        "unknown", "pending", "running", "completed", "canceled", "loading", "error"
        """
        status: JobExecutionStatus = self._job_manager.get_job_execution_status(self._id, self._machine_id)
        name, _ = betterproto.which_one_of(status, "status")
        return name

    @property
    def id(self) -> str:
        """
        Returns: The id of the job
        """
        return self._id

    @property
    def user_added(self) -> Optional[str]:
        """
        Returns: The id of the user who added the job
        """
        return self._added_user_id

    @property
    def time_added(self) -> Optional[datetime.datetime]:
        """
        Returns: The time at which the job was added
        """
        return self._time_added

    def insert_input_stream(
        self,
        name: str,
        data: List[Value],
    ) -> None:
        """Deprecated - Please use `job.push_to_input_stream`."""
        warnings.warn(
            deprecation_message(
                method="job.insert_input_stream",
                deprecated_in="1.2.0",
                removed_in="1.4.0",
                details="This method was renamed to `job.push_to_input_stream`.",
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        self.push_to_input_stream(name, data)

    def push_to_input_stream(self, name: str, data: List[Value]) -> None:
        """Push data to the input stream declared in the QUA program.
        The data is then ready to be read by the program using the advance
        input stream QUA statement.

        Multiple data entries can be inserted before the data is read by the program.

        See [Input streams](../Guides/features.md#input-streams) for more information.

        -- Available from QOP 2.0 --

        Args:
            name: The input stream name the data is to be inserted to.
            data: The data to be inserted. The data's size must match
                the size of the input stream.
        """
        if not self._capabilities.supports_input_stream:
            raise QmQuaException("`push_to_input_stream()` is not supported by the QOP version.")

        if not isinstance(data, list):
            data = [data]

        self._job_manager.insert_input_stream(self.id, name, data)



from typing import Union, TypeVar, Optional

import numpy as np
from deprecation import deprecated

from qm._loc import _get_loc
from qm.grpc import qua as _qua
from qm.exceptions import QmQuaException
from qm.serialization.expression_serializing_visitor import ExpressionSerializingVisitor
from qm.qua._type_hinting import (
    PyIntType,
    PyBoolType,
    PyArrayType,
    PyFloatType,
    PyNumberType,
    MessageVariableType,
    MessageExpressionType,
    VariableDeclarationType,
    MessageVariableOrExpression,
    fixed,
)

_ScalarExpressionType = _qua.QuaProgramAnyScalarExpression
_VariableRefType = _qua.QuaProgramVarRefExpression


def var(name: str) -> _VariableRefType:
    """A reference to a variable

    Args:
        name

    Returns:

    """
    exp = _qua.QuaProgramVarRefExpression(name=name, loc=_get_loc())
    return exp


def binary(left: _ScalarExpressionType, sop: str, right: _ScalarExpressionType) -> _ScalarExpressionType:
    """A binary operation

    Args:
        left
        sop
        right

    Returns:

    """
    if sop == "+":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.ADD
    elif sop == "-":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.SUB
    elif sop == ">":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.GT
    elif sop == "<":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.LT
    elif sop == "<=":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.LET
    elif sop == ">=":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.GET
    elif sop == "==":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.EQ
    elif sop == "*":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.MULT
    elif sop == "/":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.DIV
    elif sop == "|":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.OR
    elif sop == "&":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.AND
    elif sop == "^":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.XOR
    elif sop == "<<":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.SHL
    elif sop == ">>":
        op = _qua.QuaProgramBinaryExpressionBinaryOperator.SHR
    else:
        raise QmQuaException("Unsupported operator " + sop)

    exp = _qua.QuaProgramAnyScalarExpression(
        binary_operation=_qua.QuaProgramBinaryExpression(loc=_get_loc(), left=left, right=right, op=op)
    )
    return exp


def literal_int(value: int) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        literal=_qua.QuaProgramLiteralExpression(value=str(value), type=_qua.QuaProgramType.INT, loc=_get_loc())
    )
    return exp


def literal_bool(value: bool) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        literal=_qua.QuaProgramLiteralExpression(value=str(value), type=_qua.QuaProgramType.BOOL, loc=_get_loc())
    )
    return exp


def literal_real(value: float) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        literal=_qua.QuaProgramLiteralExpression(value=str(value), type=_qua.QuaProgramType.REAL, loc=_get_loc())
    )
    return exp


def io1() -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(variable=_qua.QuaProgramVarRefExpression(io_number=1, loc=_get_loc()))
    return exp


def io2() -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(variable=_qua.QuaProgramVarRefExpression(io_number=2, loc=_get_loc()))
    return exp


def array(
    value: _qua.QuaProgramArrayVarRefExpression,
    index_exp: Optional[_ScalarExpressionType],
) -> Union[_qua.QuaProgramArrayVarRefExpression, _ScalarExpressionType]:
    if index_exp is None:
        return value
    else:
        loc = _get_loc()
        value.loc = loc
        exp = _qua.QuaProgramAnyScalarExpression(
            array_cell=_qua.QuaProgramArrayCellRefExpression(array_var=value, index=index_exp, loc=loc)
        )
        return exp


def var_ref(value: str, index_exp: Optional[_ScalarExpressionType]) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression()

    loc = _get_loc()
    if index_exp is None:
        exp.variable = _qua.QuaProgramVarRefExpression(name=value, loc=loc)
    else:
        exp.array_cell = _qua.QuaProgramArrayCellRefExpression(
            array_var=_qua.QuaProgramArrayVarRefExpression(name=value, loc=loc),
            index=index_exp,
            loc=loc,
        )
    return exp


def lib_func(
    lib_name: str,
    func_name: str,
    *args: Union[_ScalarExpressionType, _qua.QuaProgramArrayVarRefExpression],
) -> _ScalarExpressionType:
    exp = _qua.QuaProgramAnyScalarExpression(
        lib_function=_qua.QuaProgramLibFunctionExpression(
            loc=_get_loc(), function_name=func_name, library_name=lib_name
        )
    )

    for arg in args:
        if isinstance(arg, _qua.QuaProgramArrayVarRefExpression):
            element = _qua.QuaProgramLibFunctionExpressionArgument(array=arg)
            exp.lib_function.arguments.append(element)
        else:
            element = _qua.QuaProgramLibFunctionExpressionArgument(scalar=arg)
            exp.lib_function.arguments.append(element)

    return exp


QuaExpressionType = Union["QuaExpression", MessageExpressionType]
QuaVariableType = Union["QuaVariable", MessageVariableType]

QuaGenericT = TypeVar("QuaGenericT")
QuaGenericType = Union[QuaGenericT, QuaVariableType]
QuaArrayT = TypeVar("QuaArrayT")
QuaArrayType = Union[PyArrayType[QuaArrayT], QuaVariableType]
QuaIntType = QuaGenericType[PyIntType]
QuaBoolType = QuaGenericType[PyBoolType]
QuaFloatType = QuaGenericType[PyFloatType]
QuaNumberType = QuaGenericType[PyNumberType]
QuaNumberArrayType = QuaArrayType[PyNumberType]

AllQuaTypes = Union[QuaNumberType, QuaBoolType, QuaExpressionType]
QuaGenericType = Union[QuaGenericT, QuaVariableType]


class QuaExpression:
    def __init__(self, expression: MessageVariableOrExpression):
        self._expression = expression

    def __getitem__(self, item: "QuaNumberType") -> QuaExpressionType:
        return QuaExpression(to_expression(self._expression, item))

    def unwrap(self) -> MessageVariableOrExpression:
        return self._expression

    def empty(self) -> bool:
        return self._expression is None

    def length(self) -> QuaExpressionType:
        unwrapped_element = self.unwrap()
        if isinstance(unwrapped_element, _qua.QuaProgramArrayVarRefExpression):
            array_exp = _qua.QuaProgramArrayLengthExpression(array=unwrapped_element)
            array_exp.array = unwrapped_element
            result = _qua.QuaProgramAnyScalarExpression(array_length=array_exp)
            return QuaExpression(result)
        else:
            raise QmQuaException(f"{unwrapped_element} is not an array")

    def __add__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "+", other))

    def __radd__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "+", self._expression))

    def __sub__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "-", other))

    def __rsub__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "-", self._expression))

    def __neg__(self) -> "QuaExpression":
        other = to_expression(0)
        return QuaExpression(binary(other, "-", self._expression))

    def __gt__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, ">", other))

    def __ge__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, ">=", other))

    def __lt__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "<", other))

    def __le__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "<=", other))

    def __eq__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "==", other))

    def __mul__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "*", other))

    def __rmul__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "*", self._expression))

    def __truediv__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "/", other))

    def __rtruediv__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "/", self._expression))

    def __lshift__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "<<", other))

    def __rlshift__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "<<", self._expression))

    def __rshift__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, ">>", other))

    def __rrshift__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, ">>", self._expression))

    def __and__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "&", other))

    def __rand__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "&", self._expression))

    def __or__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "|", other))

    def __ror__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "|", self._expression))

    def __xor__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(self._expression, "^", other))

    def __rxor__(self, other: "AllQuaTypes") -> "QuaExpression":
        other = to_expression(other)
        return QuaExpression(binary(other, "^", self._expression))

    def __invert__(self) -> "QuaExpression":
        other = to_expression(True)
        return QuaExpression(binary(self._expression, "^", other))

    def __str__(self) -> str:
        return ExpressionSerializingVisitor.serialize(self._expression)

    def __bool__(self):
        raise QmQuaException(
            "Attempted to use a Python logical operator on a QUA variable. If you are unsure why you got this message,"
            " please see https://qm-docs.qualang.io/guides/qua_ref#boolean-operations"
        )


class QuaVariable(QuaExpression):
    def __init__(self, expression: MessageVariableOrExpression, t: "VariableDeclarationType"):
        super().__init__(expression)
        self._type = t

    @deprecated("1.1", "1.2", details="use: '_Variable.is_fixed()' instead")
    def isFixed(self) -> bool:
        return self.is_fixed()

    @deprecated("1.1", "1.2", details="use: '_Variable.is_int()' instead")
    def isInt(self) -> bool:
        return self.is_int()

    @deprecated("1.1", "1.2", details="use: '_Variable.is_bool()' instead")
    def isBool(self) -> bool:
        return self.is_bool()

    def is_fixed(self) -> bool:
        return self._type == fixed

    def is_int(self) -> bool:
        return self._type == int

    def is_bool(self) -> bool:
        return self._type == bool


IO1 = object()
IO2 = object()


def _fix_object_data_type(obj):
    if isinstance(obj, (np.floating, np.integer, np.bool_)):
        obj_item = obj.item()
        if isinstance(obj_item, np.longdouble):
            return float(obj_item)
        else:
            return obj_item
    else:
        return obj


def to_expression(other: "AllQuaTypes", index_exp: Optional["QuaNumberType"] = None) -> MessageVariableOrExpression:
    other = _fix_object_data_type(other)
    if index_exp is not None and type(index_exp) is not _qua.QuaProgramAnyScalarExpression:
        index_exp = to_expression(index_exp, None)

    if index_exp is not None and type(other) is not _qua.QuaProgramArrayVarRefExpression:
        raise QmQuaException(f"{other} is not an array")

    if isinstance(other, QuaExpression):
        return other.unwrap()
    elif isinstance(other, _qua.QuaProgramVarRefExpression):
        return other
    elif isinstance(other, _qua.QuaProgramArrayVarRefExpression):
        return array(other, index_exp)
    elif isinstance(other, bool):  # Since bool is a subtype of int, it must be before it
        return literal_bool(other)
    elif isinstance(other, int):
        return literal_int(other)
    elif isinstance(other, float):
        return literal_real(other)
    elif other == IO1:
        return io1()
    elif other == IO2:
        return io2()
    else:
        raise QmQuaException(f"Can't handle {other}")



import datetime
from enum import Enum
from dataclasses import dataclass


class InsertDirection(Enum):
    start = 1
    end = 2


@dataclass(frozen=True)
class PendingJobData:
    job_id: str
    position_in_queue: int
    time_added: datetime.datetime
    added_by: str



from octave_sdk.octave import IFMode, ClockInfo
from octave_sdk import (
    ClockType,
    OctaveOutput,
    RFOutputMode,
    ClockFrequency,
    OctaveLOSource,
    RFInputLOSource,
    RFInputRFSource,
)

from qm.utils.deprecation_utils import throw_warning

__all__ = [
    "RFInputRFSource",
    "RFOutputMode",
    "OctaveLOSource",
    "RFInputLOSource",
    "ClockType",
    "ClockFrequency",
    "OctaveOutput",
    "IFMode",
    "ClockInfo",
]

throw_warning(
    "Octave enums should be directly imported from the octave_sdk "
    "(IFMode, and ClockInfo are imported from octave_sdk.octave), this file (qm.octave.enums)"
    "will be removed in the next version",
    category=DeprecationWarning,
)



import logging
from typing import List, Tuple, Optional

from qm.persistence import BaseStore
from qm.program.program import Program
from qm.grpc.qua_config import QuaConfig
from qm.api.frontend_api import FrontendApi
from qm.jobs.pending_job import QmPendingJob
from qm.api.models.capabilities import ServerCapabilities
from qm.api.models.compiler import CompilerOptionArguments
from qm.api.job_manager_api import create_job_manager_from_api
from qm.api.models.jobs import PendingJobData, InsertDirection
from qm.jobs.job_queue_base import QmQueueBase, JobNotFoundError
from qm.type_hinting.exceution_overrides import ExecutionOverridesType
from qm.program._execution_overrides_schema import ExecutionOverridesSchema

logger = logging.getLogger(__name__)


class QmQueue(QmQueueBase[QmPendingJob]):
    def __init__(
        self,
        config: QuaConfig,
        quantum_machine_id: str,
        frontend_api: FrontendApi,
        capabilities: ServerCapabilities,
        store: BaseStore,
    ):
        super().__init__(store, capabilities)
        self.machine_id = quantum_machine_id
        self._config = config
        self._frontend: FrontendApi = frontend_api
        self._job_manager = create_job_manager_from_api(self._frontend)

    def _get_pending_jobs(
        self, job_id: Optional[str] = None, position: Optional[int] = None, user_id: Optional[str] = None
    ) -> Tuple[QmPendingJob, ...]:
        jobs: List[PendingJobData] = self._job_manager.get_pending_jobs(self.machine_id, job_id, position, user_id)
        jobs.sort(key=lambda it: it.job_id)
        result = tuple(
            QmPendingJob(
                job_id=job.job_id,
                machine_id=self.machine_id,
                frontend_api=self._frontend,
                capabilities=self._capabilities,
                store=self._store,
            )
            for job in jobs
        )
        return result

    def add(
        self,
        program: Program,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> QmPendingJob:
        """Adds a QmJob to the queue.
        Programs in the queue will play as soon as possible.

        Args:
            program: A QUA program
            compiler_options: Optional arguments for compilation

        Example:
            ```python
            qm.queue.add(program)  # adds at the end of the queue
            qm.queue.insert(program, position)  # adds at position
            ```
        """
        if compiler_options is None:
            compiler_options = CompilerOptionArguments()
        job = self._insert(program, InsertDirection.end, compiler_options)
        return job

    def add_compiled(self, program_id: str, overrides: Optional[ExecutionOverridesType] = None) -> QmPendingJob:
        """Adds a compiled QUA program to the end of the queue, optionally
        overriding the values of analog waveforms defined in the program.
        Programs in the queue will play as soon as possible.
        For a detailed explanation see
        [Precompile Jobs](../Guides/features.md#precompile-jobs).

        Args:
            program_id: A QUA program ID returned from the compile
                function
            overrides: Object containing Waveforms to run the program
                with

        Example:
            ```python
            program_id = qm.compile(...)
            pending_job = qm.queue.add_compiled(program_id, overrides={
                'waveforms': {
                    'my_arbitrary_waveform': [0.1, 0.2, 0.3],
                    'my_constant_waveform': 0.2
                }
            })
            job = pending_job.wait_for_execution()
            ```
        """
        execution_overrides = overrides or {}

        job_id = self._frontend.add_compiled_to_queue(
            machine_id=self.machine_id,
            program_id=program_id,
            execution_overrides=ExecutionOverridesSchema().load(execution_overrides),
        )
        return QmPendingJob(
            job_id=job_id,
            machine_id=self.machine_id,
            frontend_api=self._frontend,
            capabilities=self._capabilities,
            store=self._store,
        )

    def add_to_start(
        self,
        program: Program,
        compiler_options: Optional[CompilerOptionArguments] = None,
    ) -> QmPendingJob:
        """Adds a QMJob to the start of the queue.
        Programs in the queue will play as soon as possible.

        Args:
            program: A QUA program
            compiler_options: Optional arguments for compilation

        """
        return self._insert(program, InsertDirection.start, compiler_options)

    def _insert(
        self,
        program: Program,
        insert_direction: InsertDirection,
        compiler_options: Optional[CompilerOptionArguments],
    ) -> QmPendingJob:
        """Inner function to insert a program to the queue by a given insert direction (start, end)"""
        if compiler_options is None:
            compiler_options = CompilerOptionArguments()

        job_id = self._frontend.add_to_queue(
            machine_id=self.machine_id,
            program=program.qua_program,
            compiler_options=compiler_options,
            insert_direction=insert_direction,
        )

        return QmPendingJob(
            job_id=job_id,
            machine_id=self.machine_id,
            frontend_api=self._frontend,
            capabilities=self._capabilities,
            store=self._store,
        )

    def get_at(self, position: int) -> QmPendingJob:
        """Gets the pending job object at the given position in the queue

        Args:
            position: An integer position in queue

        Returns:
            The pending job

        Example:
            ```python
            qm.queue.get(job_id)
            ```
        """
        jobs = self._get_pending_jobs(None, position)
        if len(jobs) == 0:
            raise JobNotFoundError()
        return jobs[0]

    def remove_by_id(self, job_id: str) -> int:
        """Removes the pending job object with a specific job id

        Args:
            job_id: a QMJob id

        Returns:
            The number of jobs removed

        Example:
            ```python
            qm.queue.remove_by_id(job_id)
            ```
        """
        if job_id is None or job_id == "":
            raise ValueError("job_id can not be empty")

        return self._job_manager.remove_job(
            quantum_machine_id=self.machine_id,
            job_id=job_id,
            position=None,
            user_id=None,
        )

    def remove_by_position(self, position: int) -> int:
        """Remove the PendingQmJob object by position in queue

        Args:
            position: position in queue

        Returns:
            The number of jobs removed

        Example:
            ```python
            qm.queue.remove_by_position(position)
            ```python
        """
        if position is None or position <= 0:
            raise ValueError("position must be positive")

        return self._job_manager.remove_job(
            quantum_machine_id=self.machine_id,
            job_id=None,
            position=position,
            user_id=None,
        )

    def remove_by_user_id(self, user_id: str) -> int:
        return self._job_manager.remove_job(
            quantum_machine_id=self.machine_id,
            job_id=None,
            position=None,
            user_id=user_id,
        )

    def clear(self) -> int:
        """Empties the queue from all pending jobs

        Returns:
            The number of jobs removed
        """
        return self._job_manager.remove_job(self.machine_id)



from typing import Union

import betterproto

from qm.exceptions import InvalidConfigError
from qm.grpc.qua_config import (
    QuaConfig,
    QuaConfigFemTypes,
    QuaConfigMixInputs,
    QuaConfigElementDec,
    QuaConfigControllerDec,
    QuaConfigOctoDacFemDec,
    QuaConfigPortReference,
    QuaConfigMicrowaveFemDec,
    QuaConfigAdcPortReference,
    QuaConfigDacPortReference,
)

FemTypes = Union[QuaConfigControllerDec, QuaConfigOctoDacFemDec, QuaConfigMicrowaveFemDec]


def get_fem_config_instance(fem_ref: QuaConfigFemTypes) -> FemTypes:
    _, config = betterproto.which_one_of(fem_ref, "fem_type_one_of")
    if not isinstance(config, (QuaConfigControllerDec, QuaConfigOctoDacFemDec, QuaConfigMicrowaveFemDec)):
        raise InvalidConfigError(f"FEM type {type(config)} is not supported")
    return config


def get_fem_config(
    pb_config: QuaConfig, port: Union[QuaConfigDacPortReference, QuaConfigAdcPortReference, QuaConfigPortReference]
) -> FemTypes:
    if port.controller not in pb_config.v1_beta.control_devices:
        raise InvalidConfigError("Controller not found")
    controller = pb_config.v1_beta.control_devices[port.controller]
    if port.fem not in controller.fems:
        raise InvalidConfigError("FEM not found")

    fem_ref = controller.fems[port.fem]
    config = get_fem_config_instance(fem_ref)
    return config


def element_has_mix_inputs(element: QuaConfigElementDec) -> bool:
    _, inputs_inst = betterproto.which_one_of(element, "element_inputs_one_of")
    return isinstance(inputs_inst, QuaConfigMixInputs)


def get_simulation_sampling_rate(pb_config: QuaConfig) -> float:
    """
    This function is for the simulation, for OPX1000 we simulate with 2e9Hz sampling rate,
    for OPX we simulate with 1e9Hz sampling rate. To see whether it is OPX or OPX1000, we check the config.
    """
    for controller in pb_config.v1_beta.control_devices.values():
        for fem in controller.fems.values():
            _, fem_config = betterproto.which_one_of(fem, "fem_type_one_of")
            if isinstance(fem_config, QuaConfigOctoDacFemDec):
                return 2e9
    return 1e9



COMMON_OCTAVE_PREFIX = "__oct__"


class SavedVariablesNames:
    n_lo = "n_lo"
    n_image = "n_image"
    i_scan = "i_scan"
    q_scan = "q_scan"
    lo = "lo"
    g_scan = "g_scan"
    p_scan = "p_scan"
    image = "image"


class CalibrationElementsNames:
    def __init__(self, device_name: str, channel_index: int):
        self._device_name = device_name
        self._channel_index = channel_index

    @property
    def _element_prefix(self) -> str:
        return f"{COMMON_OCTAVE_PREFIX}{self._device_name}_{self._channel_index}_"

    @property
    def iq_mixer(self) -> str:
        return self._element_prefix + "IQmixer"

    @property
    def i_offset(self) -> str:
        return self._element_prefix + "I_offset"

    @property
    def q_offset(self) -> str:
        return self._element_prefix + "Q_offset"

    @property
    def signal_analyzer(self) -> str:
        return self._element_prefix + "signal_analyzer"

    @property
    def lo_analyzer(self) -> str:
        return self._element_prefix + "lo_analyzer"

    @property
    def image_analyzer(self) -> str:
        return self._element_prefix + "image_analyzer"



from typing import Any, List, Sized, Union

from marshmallow import Schema, ValidationError, fields, post_load
from marshmallow_polyfield import PolyField  # type: ignore[import-untyped]

from qm.grpc.frontend import WaveformOverride, ExecutionOverrides
from qm.type_hinting.exceution_overrides import ExecutionOverridesType


def _non_empty(v: Sized) -> None:
    if len(v) == 0:
        raise ValidationError("List must be non-empty")


def _waveforms_schema_deserialization_disambiguation(
    data: Union[float, List[float]], _o: Any
) -> Union[fields.List, fields.Float]:
    if isinstance(data, list):
        return fields.List(fields.Float(), required=True, validate=_non_empty)
    else:
        return fields.Float(required=True)


class ExecutionOverridesSchema(Schema):
    waveforms = fields.Dict(
        keys=fields.String(),
        values=PolyField(
            deserialization_schema_selector=_waveforms_schema_deserialization_disambiguation,
            required=True,
        ),
    )

    @post_load(pass_many=False)
    def build(self, data: ExecutionOverridesType, **kwargs: Any) -> ExecutionOverrides:
        result = {}

        waveforms = data.get("waveforms", {})

        for k, v in waveforms.items():
            samples = v if isinstance(v, list) else [v]
            result[k] = WaveformOverride(samples=samples)

        return ExecutionOverrides(waveforms=result)



from qm.utils.deprecation_utils import deprecation_message
from qm.utils.protobuf_utils import LOG_LEVEL_MAP, list_fields
from qm.utils.general_utils import SERVICE_HEADER_NAME, run_until_with_timeout
from qm.utils.types_utils import (
    is_iter,
    collection_has_type,
    collection_has_type_int,
    collection_has_type_bool,
    collection_has_type_float,
    get_all_iterable_data_types,
    get_iterable_elements_datatype,
)

__all__ = [
    "LOG_LEVEL_MAP",
    "get_all_iterable_data_types",
    "collection_has_type",
    "collection_has_type_bool",
    "collection_has_type_int",
    "collection_has_type_float",
    "is_iter",
    "get_iterable_elements_datatype",
    "deprecation_message",
    "list_fields",
    "run_until_with_timeout",
    "SERVICE_HEADER_NAME",
]



import re
import logging
import warnings
import dataclasses
from typing import Any, Dict, List, Tuple, Union, Generic, TypeVar, Optional

from octave_sdk import OctaveOutput, OctaveLOSource

from qm.utils import SERVICE_HEADER_NAME
from qm.type_hinting.general import PathLike
from qm.exceptions import OctaveConnectionError
from qm.api.models.capabilities import OPX_FEM_IDX
from qm.octave.calibration_db import CalibrationDB
from qm.type_hinting.config_types import StandardPort
from qm.api.models.server_details import ConnectionDetails
from qm.octave.abstract_calibration_db import AbstractCalibrationDB

logger = logging.getLogger(__name__)

ConnectionMapping = Dict[StandardPort, Tuple[str, str]]

DEFAULT_CONNECTIONS = {
    1: "I1",
    2: "Q1",
    3: "I2",
    4: "Q2",
    5: "I3",
    6: "Q3",
    7: "I4",
    8: "Q4",
    9: "I5",
    10: "Q5",
}


PortType = TypeVar("PortType", OctaveOutput, OctaveLOSource)


@dataclasses.dataclass
class LoopbackInfo(Generic[PortType]):
    name: str
    octave_port: PortType

    def __hash__(self) -> int:
        return hash((self.name, self.octave_port))


def _convert_octave_port_to_number(port: str) -> int:
    if port.startswith(tuple("IQ")) and port.endswith(tuple("12345")) and len(port) == 2:
        return int(port[-1])
    raise ValueError(f"port {port} is not a valid octave port")


def _convert_number_to_octave_port(name: str, port: int) -> List[Tuple[str, str]]:
    return [(name, f"I{port}"), (name, f"Q{port}")]


class QmOctaveConfig:
    """
    Holds connectivity and calibrations database information
    """

    def __init__(self, fan: Any = None):
        self._devices: Dict[str, ConnectionDetails] = {}
        self._loopbacks: Optional[Dict[LoopbackInfo[OctaveLOSource], LoopbackInfo[OctaveOutput]]] = None
        self._opx_octave_port_mapping: Optional[ConnectionMapping] = None
        self._octave_to_opx_port_mapping: Optional[Dict[Tuple[str, str], StandardPort]] = None
        self._calibration_db: Optional[CalibrationDB] = None
        self._fan = fan
        self._connection_headers: Dict[str, str] = {}

    @property
    def connection_headers(self) -> Dict[str, str]:
        return self._connection_headers

    @property
    def fan(self) -> Any:
        return self._fan

    def add_device_info(
        self,
        name: str,
        host: str,
        port: int,
    ) -> None:
        """Sets the octave info  - the IP address can be either the router ip or the actual ip
        depends on the installation configuration (cluster or standalone mode)

        In order to add more than one octave run the function with different octave names ips and ports as the number of octaves in the cluster

        Args:
            name (str): The octave name
            host (str): The octave/QOP ip
            port (int): The octave port
        """
        # can be either the router ip or the actual ip (depends on what we have)
        self._devices[name] = ConnectionDetails(host, port, headers={SERVICE_HEADER_NAME: name})

    def add_device(self, name: str, details: ConnectionDetails) -> None:
        self._devices[name] = details

    def get_devices(self) -> Dict[str, ConnectionDetails]:
        return self._devices

    @property
    def devices(self) -> Dict[str, ConnectionDetails]:
        return self._devices

    def set_calibration_db(self, path: PathLike) -> None:
        """Sets the path to the calibration database

        Args:
            path (str): path to the calibration database
        """
        warnings.warn(
            "QmOctaveConfig is going to be deprecated, "
            "please set the calibration db through the QuantumMachinesManager",
            DeprecationWarning,
            stacklevel=2,
        )
        self._set_calibration_db(path)

    def _set_calibration_db(self, path: Union[PathLike, AbstractCalibrationDB]) -> None:
        if isinstance(path, AbstractCalibrationDB):
            self._calibration_db = path
        else:
            self._calibration_db = CalibrationDB(str(path))

    def set_calibration_db_without_warning(self, path: Union[PathLike, AbstractCalibrationDB]) -> None:
        self._set_calibration_db(path)

    @staticmethod
    def _check_connection_format(connection: Tuple[str, Union[int, str]]) -> bool:
        return isinstance(connection, tuple) and len(connection) == 2 and isinstance(connection[0], str)

    def set_opx_octave_mapping(self, mappings: List[Tuple[str, str]]) -> None:
        """Sets the default port mapping for each `opx, octave` names

        Will be deprecated soon, should use the "connectivity" key in "controllers" inside the configuration

        Args:
            mappings: list of tuples of [OPX_name, octave_name] to connect
        """
        warnings.warn(
            "OctaveConfig.set_opx_octave_mapping is deprecated as of 1.1.6 and will be removed in 1.3.0. "
            "The port mapping was moved to the QUA-config. Please set the mapping there.",
            DeprecationWarning,
            stacklevel=2,
        )
        if self._opx_octave_port_mapping is None:
            self._opx_octave_port_mapping = {}

        if self._opx_octave_port_mapping:
            logger.warning("Setting a new opx-octave port mapping, your configured mapping will be overridden!")

        for opx_name, octave_name in mappings:
            self._opx_octave_port_mapping.update(self.get_default_opx_octave_port_mapping(opx_name, octave_name))

    def add_opx_octave_port_mapping(
        self, connections: Dict[Union[StandardPort, Tuple[str, int]], Tuple[str, str]]
    ) -> None:
        """
        Adds port mapping which is different from the default one. should be in the form:
        ```
        {('con1', 1): ('oct1', 'I1'),
        ('con1', 2): ('oct1', 'Q1'),
        ('con1', 3): ('oct1', 'I2'),
        ('con1', 4): ('oct1', 'Q2'),
        ('con1', 5): ('oct1', 'I3'),
        ('con1', 6): ('oct1', 'Q3'),
        ('con1', 7): ('oct1', 'I4'),
        ('con1', 8): ('oct1', 'Q4'),
        ('con1', 9): ('oct1', 'I5'),
        ('con1', 10): ('oct1', 'Q5')}
        ```

        Will be deprecated soon, should use the "connectivity" key in "controllers" inside the configuration

        Args:
            connections (ConnectionMapping): mapping of OPXs to octaves connections
        """
        # ("con1", 1): ("octave1", "I1")
        # validate structure:

        warnings.warn(
            "OctaveConfig.add_opx_octave_port_mapping is deprecated as of 1.1.6 and will be removed in 1.3.0. "
            "The port mapping was moved to the QUA-config. Please set the mapping there.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self._opx_octave_port_mapping is None:
            self._opx_octave_port_mapping = {}

        standardized_connections = {}

        for key, octave_connection in connections.items():
            pattern = re.compile("([IQ])([12345])")
            if not (
                self._check_connection_format(octave_connection)
                and isinstance(octave_connection[1], str)
                and pattern.match(octave_connection[1]) is not None
            ):
                raise ValueError(
                    f"value {octave_connection} is not according to format " f'("octave_name", "octave_port")'
                )
            if len(key) == 2:
                standard_key = (key[0], OPX_FEM_IDX, key[1])
            else:
                standard_key = key
            standardized_connections[standard_key] = octave_connection
        self._opx_octave_port_mapping.update(standardized_connections)

    def add_octave_to_opx_port_mapping(
        self, connections: Dict[Tuple[str, str], Union[StandardPort, Tuple[str, int]]]
    ) -> None:
        """
        Adds port mapping from octave outputs to OPX inputs:
        ```
        {('oct1', 'I'): ('con1', 1),
        ('oct1', 'Q'): ('con1', 2)}
        ```

        Will be deprecated soon, should use the "connectivity" key in "controllers" inside the configuration

        Args:
            connections (ConnectionMapping): mapping of octaves to OPXs connections
        """

        warnings.warn(
            "OctaveConfig.add_opx_octave_port_mapping is deprecated as of 1.1.6 and will be removed in 1.3.0. "
            "The port mapping was moved to the QUA-config. Please set the mapping there.",
            DeprecationWarning,
            stacklevel=2,
        )

        if self._octave_to_opx_port_mapping is None:
            self._octave_to_opx_port_mapping = {}

        standardized_connections = {}

        for key, opx_connection in connections.items():

            if len(opx_connection) == 2:
                standard_key = (opx_connection[0], OPX_FEM_IDX, opx_connection[1])
            else:
                standard_key = opx_connection
            standardized_connections[key] = standard_key
        self._octave_to_opx_port_mapping.update(standardized_connections)

    def get_opx_octave_port_mapping(self) -> ConnectionMapping:
        """Get the configured opx-octave connections

        Returns:
            Mapping of the configured OPXs to octaves connections
        """
        if self._opx_octave_port_mapping is None:
            return {}
        warnings.warn(
            "Setting the mapping was move to the QUA-config. "
            "It is deprecated since 1.1.6 and will be removed in 1.3.0."
            "Please set the mapping there.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._opx_octave_port_mapping

    def get_octave_to_opx_port_mapping(self) -> Dict[Tuple[str, str], StandardPort]:
        """Get the configured octave-opx connections

        Returns:
            Mapping of the configured octaves to OPXs connections
        """
        if self._octave_to_opx_port_mapping is None:
            return {}
        warnings.warn(
            "Setting the mapping was move to the QUA-config. "
            "It is deprecated since 1.2.2 and will be removed in 1.3.0."
            "Please set the mapping there.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._octave_to_opx_port_mapping

    @staticmethod
    def get_default_opx_octave_port_mapping(controller_name: str, octave_name: str) -> ConnectionMapping:
        """
        Get the default opx-octave connections

        Args:
            controller_name (str): OPX name
            octave_name (str): octave name

        Returns:
            Mapping of the given OPX to the given octave
        """
        default_connections = {}
        for opx_port, octave_port in DEFAULT_CONNECTIONS.items():
            default_connections[(controller_name, OPX_FEM_IDX, opx_port)] = (octave_name, octave_port)
        return default_connections

    @property
    def calibration_db(self) -> Optional[CalibrationDB]:
        return self._calibration_db

    def add_lo_loopback(
        self,
        octave_output_name: str,
        octave_output_port: OctaveOutput,
        octave_input_name: str,
        octave_input_port: OctaveLOSource,
    ) -> None:
        """Adds a loopback between an OctaveOutput and an OctaveLOSource

        Args:
            octave_output_name (str): octave name
            octave_output_port (OctaveOutput): octave output port
                according to OctaveOutput
            octave_input_name (str): octave name
            octave_input_port (OctaveLOSource): the LO input port
                according to OctaveLOSource
        """
        warnings.warn(
            "OctaveConfig.add_lo_loopback is deprecated as of 1.1.6 and will be removed in 1.3.0. "
            "Please set the loopbacks there.",
            DeprecationWarning,
            stacklevel=2,
        )
        if self._loopbacks is None:
            self._loopbacks = {}

        if octave_output_name != octave_input_name:
            raise ValueError("lo loopback between different octave devices are not supported")
        loop_back_source = LoopbackInfo(octave_output_name, octave_output_port)
        loop_back_destination = LoopbackInfo(octave_input_name, octave_input_port)
        self._loopbacks[loop_back_destination] = loop_back_source

    def get_lo_loopbacks_by_octave(self, octave_name: str) -> Dict[OctaveLOSource, OctaveOutput]:
        """
        Gets a list of all loop backs by octave name

        Args:
            octave_name (str): octave name to get LO loopback for

        Returns:
            Dictionary with all the LO loopbacks
        """
        if self._loopbacks is None:
            return {}

        warnings.warn(
            "Loopbacks were moved to the QUA-config, this function will be removed in 1.3.0,"
            "please set the loopbacks there, and get them from there",
            DeprecationWarning,
            stacklevel=2,
        )
        result = {}
        for destination, source in self._loopbacks.items():
            if destination.name == octave_name:
                result[destination.octave_port] = source.octave_port
        return result

    def get_opx_iq_ports(self, octave_output_port: Tuple[str, int]) -> Tuple[StandardPort, StandardPort]:
        conns = self.get_opx_octave_port_mapping()
        inv_conns = {v: k for k, v in conns.items()}
        octave_input_port_i, octave_input_port_q = _convert_number_to_octave_port(
            octave_output_port[0], octave_output_port[1]
        )
        if octave_input_port_i not in inv_conns:
            octave_name, octave_port = octave_input_port_i
            raise KeyError(f"Could not find opx connections to port " f"'{octave_port}' of octave '{octave_name}'")

        if octave_input_port_q not in inv_conns:
            octave_name, octave_port = octave_input_port_q
            raise KeyError(f"Could not find opx connections to port " f"'{octave_port}' of octave '{octave_name}'")
        return inv_conns[octave_input_port_i], inv_conns[octave_input_port_q]

    def get_octave_input_port(
        self,
        opx_i_port: StandardPort,
        opx_q_port: StandardPort,
    ) -> Optional[Tuple[str, int]]:
        # check both ports are going to the same mixer
        connections = self.get_opx_octave_port_mapping()
        i_octave_port = connections.get(opx_i_port)
        q_octave_port = connections.get(opx_q_port)

        if i_octave_port is None and q_octave_port is None:
            return None

        if i_octave_port is None and q_octave_port is not None:
            raise OctaveConnectionError("I port is not connected to any octave and Q port is connected to an octave")
        if i_octave_port is not None and q_octave_port is None:
            raise OctaveConnectionError("Q port is not connected to any octave and I port is connected to an octave")
        assert i_octave_port is not None
        assert q_octave_port is not None
        if i_octave_port[0] != q_octave_port[0]:
            raise OctaveConnectionError("I and Q are not connected to the same octave")
        if i_octave_port[1][-1] != q_octave_port[1][-1]:
            raise OctaveConnectionError("I and Q are not connected to the same octave input")

        return i_octave_port[0], _convert_octave_port_to_number(i_octave_port[1])



from typing import Dict, List
from dataclasses import dataclass

import betterproto
import numpy as np
import numpy.typing as nt

from qm.grpc import results_analyser
from qm.grpc.results_analyser import (
    ProgramStreamMetadata,
    IterationDataForIntIterationValues,
    IterationDataForDoubleIterationValues,
    IterationDataForEachIntIterationValues,
    IterationDataForEachDoubleIterationValues,
)


@dataclass
class IterationData:
    iteration_variable_name: str
    iteration_values: nt.NDArray[np.generic]


@dataclass
class StreamMetadataError:
    error: str
    location: str


@dataclass
class StreamMetadata:
    stream_name: str
    iteration_values: List[IterationData]


def _get_numpy_array_from_proto_iteration_data(
    iteration_data: results_analyser.IterationData,
) -> nt.NDArray[np.generic]:
    name, value = betterproto.which_one_of(iteration_data, "iterationValues")
    if name in ("for_each_int_iteration_values", "for_each_double_iteration_values"):
        assert isinstance(value, (IterationDataForEachDoubleIterationValues, IterationDataForEachIntIterationValues))
        return np.array(value.values)

    assert isinstance(value, (IterationDataForDoubleIterationValues, IterationDataForIntIterationValues))
    start = value.start_value
    step = value.step
    stop = start + step * value.number_of_iterations

    if name == "for_double_iteration_values":
        stop = round(stop, 9)

    return np.arange(start=start, step=step, stop=stop)


def _get_stream_metadata_dict_from_proto_resp(
    program_metadata: ProgramStreamMetadata,
) -> Dict[str, StreamMetadata]:
    stream_metadata_dict = {
        x.stream_name: StreamMetadata(
            x.stream_name,
            [
                IterationData(
                    y.iteration_variable_name,
                    _get_numpy_array_from_proto_iteration_data(y),
                )
                for y in x.iteration_data
            ],
        )
        for x in program_metadata.stream_metadata
    }

    return stream_metadata_dict



from qm.api.models.server_details import ConnectionDetails
from qm.api.v2.job_api.generic_apis import ElementGenericApi
from qm.grpc.v2 import (
    GetDigitalDelayRequest,
    SetDigitalDelayRequest,
    GetDigitalBufferRequest,
    GetInputDcOffsetRequest,
    SetDigitalBufferRequest,
    SetInputDcOffsetRequest,
    SetOscillatorFrequencyRequest,
    SetOscillatorFrequencyRequestUpdateComponentSelection,
)


class ElementPortApi(ElementGenericApi):
    def __init__(self, connection_details: ConnectionDetails, job_id: str, element_id: str, port_name: str) -> None:
        super().__init__(connection_details, job_id, element_id)
        self._port = port_name


class AnalogOutputApi(ElementPortApi):
    def set_dc_offset(self, value: float) -> None:
        request = SetInputDcOffsetRequest(job_id=self._id, qe=self._element_id, port=self._port, offset=value)
        self._run(self._stub.set_input_dc_offset(request, timeout=self._timeout))

    def get_dc_offset(self) -> float:
        request = GetInputDcOffsetRequest(job_id=self._id, qe=self._element_id, port=self._port)
        response = self._run(self._stub.get_input_dc_offset(request, timeout=self._timeout))
        return response.offset


class MwOutputApi(ElementGenericApi):
    def set_oscillator_frequency(self, frequency: float) -> None:
        request = SetOscillatorFrequencyRequest(
            job_id=self._id,
            qe=self._element_id,
            new_frequency_hz=frequency,
            update_component=SetOscillatorFrequencyRequestUpdateComponentSelection.downconverter,
        )
        self._run(self._stub.set_oscillator_frequency(request, timeout=self._timeout))


class DigitalInputApi(ElementPortApi):
    def set_delay(self, value: int) -> None:
        request = SetDigitalDelayRequest(job_id=self._id, qe=self._element_id, port=self._port, delay=value)
        self._run(self._stub.set_digital_delay(request, timeout=self._timeout))

    def get_delay(self) -> int:
        request = GetDigitalDelayRequest(job_id=self._id, qe=self._element_id, port=self._port)
        response = self._run(self._stub.get_digital_delay(request, timeout=self._timeout))
        return response.delay

    def set_buffer(self, value: int) -> None:
        request = SetDigitalBufferRequest(job_id=self._id, qe=self._element_id, port=self._port, buffer=value)
        self._run(self._stub.set_digital_buffer(request, timeout=self._timeout))

    def get_buffer(self) -> int:
        request = GetDigitalBufferRequest(job_id=self._id, qe=self._element_id, port=self._port)
        response = self._run(self._stub.get_digital_buffer(request, timeout=self._timeout))
        return response.buffer



import time
import logging
from pathlib import Path
from dataclasses import asdict, dataclass
from typing import Dict, Tuple, Union, Mapping, Optional, cast

from tinydb.table import Document
from tinydb import Query, TinyDB, where
from tinydb.storages import JSONStorage

from qm.octave.octave_mixer_calibration import MixerCalibrationResults
from qm.octave.calibration_utils import Correction, convert_to_correction
from qm.octave.abstract_calibration_db import AbstractCalibrationDB, AbstractIFCalibration, AbstractLOCalibration

logger = logging.getLogger(__name__)


@dataclass
class _Mode:
    octave_name: str
    octave_channel: int


@dataclass
class _LOMode:
    mode_id: int
    lo_freq: Union[int, float]
    gain: Optional[float]
    latest: int


@dataclass
class _IFMode:
    lo_mode_id: int
    if_freq: float
    latest: int


@dataclass
class LOCalibrationDBSchema(AbstractLOCalibration):
    i0: float
    q0: float
    dc_gain: float
    dc_phase: float
    temperature: float
    timestamp: float
    method: str

    def get_i0(self) -> float:
        return self.i0

    def get_q0(self) -> float:
        return self.q0


@dataclass
class IFCalibrationDBSchema(AbstractIFCalibration):
    gain: float
    phase: float
    temperature: float
    timestamp: float
    method: str

    def get_correction(self) -> Correction:
        return convert_to_correction(self.gain, self.phase)


class _ModeNotFoundError(KeyError):
    def __init__(self, query: Tuple[Union[str, Union[int, float]], ...]):
        self._query = query

    def __str__(self) -> str:
        return f"Didn't find mode for {self._query}"


class CalibrationDB(AbstractCalibrationDB):
    """
    Class for the calibration database.
    """

    def __init__(self, path: Union[Path, str]) -> None:
        self._file_path: Path = Path(path) / "calibration_db.json"
        self._db = TinyDB(self._file_path, indent=4, separators=(",", ": "), storage=JSONStorage)

    def __del__(self) -> None:
        self._db.close()

    def reset(self) -> None:
        self._db.close()

        self._file_path.unlink()
        self._db = TinyDB(self._file_path, indent=4, separators=(",", ": "), storage=JSONStorage)

    def _query_mode(self, octave_channel: Tuple[str, int]) -> Optional[Document]:
        return cast(
            Optional[Document],
            self._db.table("modes").get(
                (Query().octave_name == octave_channel[0]) & (Query().octave_channel == octave_channel[1])
            ),
        )

    def _get_timestamp(self, doc: Document) -> float:
        a = cast(Document, self._db.table("lo_cal").get(doc_id=doc["latest"]))
        return cast(float, a["timestamp"])

    def _query_lo_mode(self, mode_id: int, lo_freq: Union[int, float], gain: Optional[float]) -> Optional[Document]:
        if gain is not None:
            return cast(
                Optional[Document],
                self._db.table("lo_modes").get(
                    (Query().mode_id == mode_id) & (Query().lo_freq == lo_freq) & (Query().gain == gain)
                ),
            )

        lo_modes = self._db.table("lo_modes").search((Query().mode_id == mode_id) & (Query().lo_freq == lo_freq))
        if not lo_modes:
            return None

        return max(lo_modes, key=self._get_timestamp)

    def _query_if_mode(self, lo_mode_id: int, if_freq: float) -> Optional[Document]:
        return cast(
            Optional[Document],
            self._db.table("if_modes").get((Query().lo_mode_id == lo_mode_id) & (Query().if_freq == if_freq)),
        )

    def _mode_id(self, octave_channel: Tuple[str, int], create: bool = False) -> int:
        query_result = self._query_mode(octave_channel)
        if query_result is None:
            if create:
                return self._db.table("modes").insert(asdict(_Mode(*octave_channel)))
            raise _ModeNotFoundError(octave_channel)
        else:
            return query_result.doc_id

    def _lo_mode_id(self, mode_id: int, lo_freq: Union[int, float], gain: Optional[float], create: bool = False) -> int:
        query_result = self._query_lo_mode(mode_id, lo_freq, gain)
        if query_result is None:
            if create:
                return self._db.table("lo_modes").insert(asdict(_LOMode(mode_id, lo_freq, gain, 0)))
            else:
                raise _ModeNotFoundError((mode_id, lo_freq))
        else:
            return query_result.doc_id

    def _if_mode_id(self, lo_mode_id: int, if_freq: float, create: bool = False) -> int:
        query_result = self._query_if_mode(lo_mode_id, if_freq)
        if query_result is None:
            if create:
                return self._db.table("if_modes").insert(asdict(_IFMode(lo_mode_id, if_freq, 0)))
            else:
                raise _ModeNotFoundError((lo_mode_id, if_freq))
        else:
            return query_result.doc_id

    def _update_lo_calibration(
        self,
        octave_channel: Tuple[str, int],
        lo_freq: Union[int, float],
        gain: Optional[float],
        i0: float,
        q0: float,
        dc_gain: float,
        dc_phase: float,
        temperature: float,
        method: str = "",
    ) -> None:

        mode_id = self._mode_id(octave_channel, create=True)
        lo_mode_id = self._lo_mode_id(mode_id, lo_freq, gain, create=True)

        timestamp = time.time()

        lo_cal_id = self._db.table("lo_cal").insert(
            asdict(LOCalibrationDBSchema(i0, q0, dc_gain, dc_phase, temperature, timestamp, method))
        )

        self._db.table("lo_modes").update(asdict(_LOMode(mode_id, lo_freq, gain, lo_cal_id)), doc_ids=[lo_mode_id])

    def _update_if_calibration(
        self,
        octave_channel: Tuple[str, int],
        lo_freq: Union[int, float],
        output_gain: Optional[float],
        if_freq: float,
        gain: float,
        phase: float,
        temperature: float,
        method: str = "",
    ) -> None:

        mode_id = self._mode_id(octave_channel, create=True)
        lo_mode_id = self._lo_mode_id(mode_id, lo_freq, output_gain, create=True)
        if_mode_id = self._if_mode_id(lo_mode_id, if_freq, create=True)

        timestamp = time.time()
        if_cal_id = self._db.table("if_cal").insert(
            asdict(IFCalibrationDBSchema(gain, phase, temperature, timestamp, method))
        )

        self._db.table("if_modes").update(asdict(_IFMode(lo_mode_id, if_freq, if_cal_id)), doc_ids=[if_mode_id])

    def get_lo_cal(
        self, octave_channel: Tuple[str, int], lo_freq: Union[int, float], gain: Optional[float]
    ) -> Optional[AbstractLOCalibration]:
        """
        Get the LO calibration for a given octave channel, LO frequency, and gain
        Args:
            octave_channel: The octave channel to get the calibration for.
            lo_freq: The LO frequency to get the calibration for.
            gain: The gain to get the calibration for.
        Returns:
            The LO calibration for the given parameters, or None if it doesn't exist.
        """
        try:
            mode_id = self._mode_id(octave_channel)
        except _ModeNotFoundError:
            return None
        lo_mode = self._query_lo_mode(mode_id, lo_freq, gain)
        if lo_mode is None or lo_mode["latest"] == 0:
            return None

        lo_cal = cast(Optional[Document], self._db.table("lo_cal").get(doc_id=lo_mode["latest"]))
        if lo_cal is None:
            return None
        lo_cal.pop("lo_mode_id", None)
        lo_cal_obj = LOCalibrationDBSchema(**lo_cal)

        return lo_cal_obj

    def get_if_cal(
        self,
        octave_channel: Tuple[str, int],
        lo_freq: Union[int, float],
        gain: Optional[float],
        if_freq: Union[int, float],
    ) -> Optional[IFCalibrationDBSchema]:
        """
        Get the IF calibration for a given octave channel, LO frequency, gain, and IF frequency.
        Args:
            octave_channel: The octave channel to get the calibration for.
            lo_freq: The LO frequency to get the calibration for.
            gain: The gain to get the calibration for.
            if_freq: The IF frequency to get the calibration for.
        Returns:
            The IF calibration for the given parameters, or None if it doesn't exist.
        """
        try:
            mode_id = self._mode_id(octave_channel, create=False)
            lo_mode_id = self._lo_mode_id(mode_id, lo_freq, gain, create=False)
            if_mode = self._query_if_mode(lo_mode_id, if_freq)
        except _ModeNotFoundError:
            return None

        if if_mode is None or if_mode["latest"] == 0:
            return None

        if_cal = self._db.table("if_cal").get(doc_id=if_mode["latest"])
        if if_cal is None:
            return None
        assert isinstance(if_cal, dict)
        if_cal.pop("if_mode_id", None)
        if_cal_obj = IFCalibrationDBSchema(**if_cal)

        return if_cal_obj

    def get_all_if_cal_for_lo(
        self, octave_channel: Tuple[str, int], lo_freq: Union[int, float], gain: Optional[float]
    ) -> Mapping[Union[int, float], IFCalibrationDBSchema]:
        """
        Get IF calibration for all IF frequencies for a given octave channel, LO frequency, and gain.
        Args:
            octave_channel: The octave channel to get the calibration for.
            lo_freq: The LO frequency to get the calibration for.
            gain: The gain to get the calibration for.
        Returns:
            A dictionary of IF frequencies to their respective IF calibration.
        """
        try:
            mode_id = self._mode_id(octave_channel)
            lo_mode_id = self._lo_mode_id(mode_id, lo_freq, gain)
        except _ModeNotFoundError:
            return {}

        if_modes = self._db.table("if_modes").search(where("lo_mode_id") == lo_mode_id)

        if if_modes is None:
            return {}

        if_dict: Dict[Union[int, float], IFCalibrationDBSchema] = {}
        for if_mode in if_modes:
            if_freq = if_mode["if_freq"]
            if_cal = self.get_if_cal(octave_channel, lo_freq, gain, if_freq)

            if if_cal:
                if_dict[if_freq] = if_cal

        return if_dict

    def update_calibration_result(
        self, result: MixerCalibrationResults, octave_channel: Tuple[str, int], method: str = ""
    ) -> None:
        """
        Update the calibration database with the given calibration results.
        Args:
            result: The calibration results to update the database with.
            octave_channel: The octave channel to update the database with.
            method: Deprecated.
        """

        for (lo_freq, output_gain), lo_cal in result.items():
            self._update_lo_calibration(
                octave_channel,
                lo_freq,
                output_gain,
                lo_cal.i0,
                lo_cal.q0,
                lo_cal.dc_gain,
                lo_cal.dc_phase,
                lo_cal.temperature,
                method,
            )

            for if_freq, if_cal in lo_cal.image.items():
                fine_cal = if_cal.fine
                self._update_if_calibration(
                    octave_channel,
                    lo_freq,
                    output_gain,
                    if_freq,
                    fine_cal.gain,
                    fine_cal.phase,
                    lo_cal.temperature,
                    method,
                )
